###############################################################################
#
#   Green17: extinction model from Green et al. (2017)
#
###############################################################################
import os, os.path
import numpy
import h5py
from mwdust.util.download import dust_dir, downloader
from mwdust.HierarchicalHealpixMap import HierarchicalHealpixMap
_DEGTORAD= numpy.pi/180.
_greendir= os.path.join(dust_dir, 'green17')
class Green17(HierarchicalHealpixMap):
    """extinction model from Green et al. (2018)"""
    def __init__(self,filter=None,sf10=True,load_samples=False,
                 interpk=1):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the Green et al. (2017) dust map
	   The reddening vector is not the one used in Green et al. (2015)
	   But instead: Schlafly et al. (2016)
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
           load_samples= (False) if True, also load the samples
           interpk= (1) interpolation order
        OUTPUT:
           object
        HISTORY:
           2019-10-09 - Adopted - Rybizki (MPIA)
        """
        HierarchicalHealpixMap.__init__(self,filter=filter,sf10=sf10)
        #Read the map
        with h5py.File(os.path.join(_greendir,'bayestar2017.h5'),'r') \
                as greendata:
            self._pix_info= greendata['/pixel_info'][:]
            if load_samples:
                self._samples= greendata['/samples'][:]
            self._best_fit= greendata['/best_fit'][:]
            self._GR= greendata['/GRDiagnostic'][:]
        # Utilities
        self._distmods= numpy.linspace(4,19,31)
        self._minnside= numpy.amin(self._pix_info['nside'])
        self._maxnside= numpy.amax(self._pix_info['nside'])
        nlevels= int(numpy.log2(self._maxnside//self._minnside))+1
        self._nsides= [self._maxnside//2**ii for ii in range(nlevels)]
        self._indexArray= numpy.arange(len(self._pix_info['healpix_index']))
        # For the interpolation
        self._intps= numpy.zeros(len(self._pix_info['healpix_index']),
                                 dtype='object') #array to cache interpolated extinctions
        self._interpk= interpk
        return None

    def substitute_sample(self,samplenum):
        """
        NAME:
           substitute_sample
        PURPOSE:
           substitute a sample for the best fit to get the extinction from a sample with the same tools; need to have setup the instance with load_samples=True
        INPUT:
           samplenum - sample's index to load
        OUTPUT:
           (none; just resets the instance to use the sample rather than the best fit; one cannot go back to the best fit after this))
        HISTORY:
           2019-10-09 - Adopted - Rybizki (MPIA)
        """
        # Substitute the sample
        self._best_fit= self._samples[:,samplenum,:]
        # Reset the cache
        self._intps= numpy.zeros(len(self._pix_info['healpix_index']),
                                 dtype='object') #array to cache interpolated extinctions
        return None

    @classmethod
    def download(cls, test=False):
        # Download Green et al. 2018 PanSTARRS data
        green17_path = os.path.join(dust_dir, "green17", "bayestar2017.h5")
        if not os.path.exists(green17_path):
            if not os.path.exists(os.path.join(dust_dir, "green17")):
                os.mkdir(os.path.join(dust_dir, "green17"))
            _GREEN17_URL = "https://dataverse.harvard.edu/api/access/datafile/:persistentId?persistentId=doi:10.7910/DVN/LCYHJG/S7MP4P"
            downloader(_GREEN17_URL, green17_path, cls.__name__, test=test)
        return None



###############################################################################
#
#   Combined19: extinction model obtained from a combination of Marshall et al.
#               (2006), Green et al. (2019), and Drimmel et al. (2003)
#
###############################################################################
import os, os.path
import numpy
import h5py
from mwdust.util.download import dust_dir, downloader
from mwdust.HierarchicalHealpixMap import HierarchicalHealpixMap
_DEGTORAD= numpy.pi/180.
_combineddir= os.path.join(dust_dir,'combined19')
class Combined19(HierarchicalHealpixMap):
    """extinction model obtained from a combination of Marshall et al.
    (2006), Green et al. (2019), and Drimmel et al. (2003)"""
    def __init__(self,filter=None,sf10=True,load_samples=False,
                 interpk=1):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the combined dust map
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
           interpk= (1) interpolation order
        OUTPUT:
           object
        HISTORY:
           2015-07-28 - Started - Bovy (UofT)
           2019-09-12 - Updated - Rybizki (MPIA)	   
        """
        HierarchicalHealpixMap.__init__(self,filter=filter,sf10=sf10)
        #Read the map
        with h5py.File(os.path.join(_combineddir,'combine19.h5'),'r') \
                as combineddata:
            self._pix_info= combineddata['/pixel_info'][:]
            self._best_fit= combineddata['/best_fit'][:]
        # Utilities
        self._distmods= numpy.linspace(4,18.875,120)
        self._minnside= numpy.amin(self._pix_info['nside'])
        self._maxnside= numpy.amax(self._pix_info['nside'])
        nlevels= int(numpy.log2(self._maxnside//self._minnside))+1
        self._nsides= [self._maxnside//2**ii for ii in range(nlevels)]
        self._indexArray= numpy.arange(len(self._pix_info['healpix_index']))
        # For the interpolation
        self._intps= numpy.zeros(len(self._pix_info['healpix_index']),
                                 dtype='object') #array to cache interpolated extinctions
        self._interpk= interpk
        return None

    @classmethod
    def download(cls, test=False):
        # Download the combined map: Marshall+Green19+Drimmel for full sky coverage
        combined19_path = os.path.join(dust_dir, "combined19", "combine19.h5")
        if not os.path.exists(combined19_path):
            if not os.path.exists(os.path.join(dust_dir, "combined19")):
                os.mkdir(os.path.join(dust_dir, "combined19"))
            _COMBINED19_URL = "https://zenodo.org/record/3566060/files/combine19.h5"
            downloader(_COMBINED19_URL, combined19_path, cls.__name__, test=test)
        return None



import re
import os, os.path
import numpy

#A / E(B-v)
avebv=  {}
avebvsf= {}
def _read_extCurves():
    """Read the extinction curves files in extCurves"""
    extFile= open(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                               'extCurves','extinction.tbl'),'r')
    global avebv
    global avebvsf
    for line in extFile:
        if line[0] == '\\': continue
        if line[0] == '|': continue
        vals= re.split('\s\s+',line)
        avebv[vals[0].strip()]= float(vals[4])
        avebvsf[vals[0].strip()]= float(vals[2])    
    # Add filters from Schlafly & Finkbeiner
    extFile= open(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                               'extCurves','apj398709t6_ascii.txt'),'r')
    cnt= 0
    for line in extFile:
        cnt+= 1
        if cnt < 6: continue
        if cnt > 49: break
        vals= line.split()
        # Each line has 2 filters
        filter1= '%s %s' % (vals[0],vals[1])
        if not filter1 in avebv:
            avebv[filter1]= numpy.nan
            avebvsf[filter1]= float(vals[4])
        filter2= '%s %s' % (vals[7],vals[8])
        if not filter2 in avebv:
            avebv[filter2]= numpy.nan
            avebvsf[filter2]= float(vals[11])

_read_extCurves()

def aebv(filter,sf10=True):
    """
    NAME:
       aebv
    PURPOSE:
       return A_filter / E(B-V), necessary to turn SFD E(B-V) into total extinctions
    INPUT:
       filter - filter to use (e.g., '2MASS Ks')
       sf10= (True) if True, use the values from Schlafly & Finkbeiner 2010, which use an updated extinction law, source spectrum, and recalibrated SFD map
    OUTPUT:
       A_filter / E(B-V)
    HISTORY:
       2013-11-24 - Written - Bovy (IAS)
    """
    if filter.lower() == 'e(b-v)':
        return 0.884
    if sf10:
        if not filter in avebvsf:
            raise ValueError("Requested filter is not supported")
        else:
            return avebvsf[filter]
    else:
        if not filter in avebv:
            raise ValueError("Requested filter is not supported")
        return avebv[filter]



import pathlib
import numpy
from numpy.random import default_rng
rng= default_rng()

def test_distance_independent():
    # Test that the SFD extinction is indepdendent of distance 
    # for a given line of sight.
    from mwdust import SFD
    sfd= SFD()
    glons= rng.uniform(0.,360.,size=20)
    glats= rng.uniform(-90.,90.,size=20)
    dists= rng.uniform(0.1,10.,size=5)
    ebvs= numpy.array([[sfd(glons[ii],glats[ii],dists[jj]) \
                            for ii in range(len(glons))]
                            for jj in range(len(dists))]).T     
    assert numpy.all(numpy.fabs(ebvs-ebvs[0]) < 10.**-10.), \
        'SFD extinction is not independent of distance for a given line of sight'
    return None

def test_against_known_values():
    # Test that the SFD extinction without interpolation agrees with a table of known values
    # These were computed using mwdust on an M1 mac, so this test is really
    # of consistency between other OSs and architectures
    from mwdust import SFD
    sfd= SFD(interp=False)
    known= numpy.loadtxt(pathlib.Path(__file__).parent / 'sfd_benchmark.dat',delimiter=',')
    glons= known.T[0]
    glats= known.T[1]
    ebvs= known.T[2]
    assert numpy.all(ebvs-[sfd(glon,glat,1.)[0]
                            for glon,glat in zip(glons,glats)] < 10.**-8.), \
        'SFD extinction without interpolation does not agree with known values'

    # Test that the SFD extinction with interpolation agrees with a table of known values
    # These were computed using mwdust on a linux server, same cavert above
    from mwdust import SFD
    sfd= SFD()
    # specifically contains lb = (30, 3), a previously known inconsistency between Linux and Windows
    known= numpy.loadtxt(pathlib.Path(__file__).parent / 'sfd_mwdust12_linux_benchmark.dat',delimiter=',')
    glons= known.T[0]
    glats= known.T[1]
    ebvs= known.T[2]
    assert numpy.all(ebvs-[sfd(glon,glat,1.)[0]
                            for glon,glat in zip(glons,glats)] < 10.**-8.), \
        'SFD extinction does not agree with known values'
    return None



import os
import pathlib
import shutil
import requests
import tqdm

dust_dir = os.environ.get("DUST_DIR")
if dust_dir is None:
    dust_dir = os.path.expanduser(os.path.join("~", ".mwdust"))
if not os.path.exists(dust_dir):
    os.mkdir(dust_dir)

def downloader(url, fullfilename, name, test=False):
    """
    url: URL of data file
    fullfilename: full local path
    name: name of the task
    """
    user_agent = "Mozilla/5.0"
    r = requests.get(url, stream=True, allow_redirects=True, verify=True, headers={"User-Agent": user_agent})
    if r.status_code == 404:
        raise ConnectionError(f"Cannot find {name} data file at {url}")
    r.raise_for_status()  # Will only raise for 4xx codes
    if not test:
        file_size = int(r.headers.get('Content-Length', 0))
        path = pathlib.Path(fullfilename).expanduser().resolve()
        path.parent.mkdir(parents=True, exist_ok=True)

        # r.raw.read
        with tqdm.tqdm.wrapattr(r.raw, "read", total=file_size, desc=f"Download {name}") as r_raw:
            with path.open("wb") as f:
                shutil.copyfileobj(r_raw, f)



###############################################################################
#
#   Marshall06: extinction model from Marshall et al. 2006 2006A&A...453..635M
#
###############################################################################
import os, os.path
import sys
import gzip
import numpy
from scipy import interpolate
from astropy.io import ascii
from mwdust.util.extCurves import aebv
from mwdust.util.tools import cos_sphere_dist
from mwdust.util.download import dust_dir, downloader
from mwdust.DustMap3D import DustMap3D

try:
    from galpy.util import plot as bovy_plot
    _BOVY_PLOT_LOADED= True
except ImportError:
    _BOVY_PLOT_LOADED= False
from matplotlib import pyplot
_DEGTORAD= numpy.pi/180.
_marshalldir= os.path.join(dust_dir,'marshall06')
_ERASESTR= "                                                                                "
class Marshall06(DustMap3D):
    """extinction model from Marshall et al. 2006 2006A&A...453..635M"""
    def __init__(self,filter=None,sf10=True):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the Marshall06 dust map
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
        OUTPUT:
           object
        HISTORY:
           2013-11-24 - Started - Bovy (IAS)
        """
        DustMap3D.__init__(self,filter=filter)
        self._sf10= sf10
        #Read the maps
        sys.stdout.write('\r'+"Reading Marshall et al. (2006) data file ...\r")
        sys.stdout.flush()
        self._marshalldata= ascii.read(os.path.join(_marshalldir,
                                                         'table1.dat'),
                                            readme=os.path.join(_marshalldir,
                                                                'ReadMe'),
                                            guess=False, format='cds',
                                            fill_values=[('', '-999')])
        sys.stdout.write('\r'+_ERASESTR+'\r')
        sys.stdout.flush()
        #Sort the data on l and then b
        negIndx= self._marshalldata['GLON'] > 180.
        self._marshalldata['GLON'][negIndx]= self._marshalldata['GLON'][negIndx]-360.
        sortIndx= numpy.arange(len(self._marshalldata))
        keyArray= (self._marshalldata['GLON']+self._marshalldata['GLAT']/100.).data
        sortIndx= sorted(sortIndx,key=lambda x: keyArray[x])
        self._marshalldata= self._marshalldata[sortIndx]
        self._dl= 0.25
        self._db= 0.25
        self._intps= numpy.zeros(len(self._marshalldata),dtype='object') #array to cache interpolated extinctions
        return None

    def _evaluate(self,l,b,d):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the dust-map
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
           d- distance (kpc) can be array
        OUTPUT:
           extinction
        HISTORY:
           2013-12-12 - Started - Bovy (IAS)
        """
        if isinstance(l,numpy.ndarray) or isinstance(b,numpy.ndarray):
            raise NotImplementedError("array input for l and b for Drimmel dust map not implemented")
        lbIndx= self._lbIndx(l,b)
        if self._intps[lbIndx] != 0:
            out= self._intps[lbIndx](d)
        else:
            tlbData= self.lbData(l,b,addBC=True)
            interpData=\
                interpolate.InterpolatedUnivariateSpline(tlbData['dist'],
                                                         tlbData['aks'],
                                                         k=1)
            out= interpData(d)
            self._intps[lbIndx]= interpData
        if self._filter is None:
            return out/aebv('2MASS Ks',sf10=self._sf10)
        else:
            return out/aebv('2MASS Ks',sf10=self._sf10)\
                *aebv(self._filter,sf10=self._sf10)

    def dust_vals_disk(self,lcen,bcen,dist,radius):
        """
        NAME:
           dust_vals_disk
        PURPOSE:
           return the distribution of extinction within a small disk as samples
        INPUT:
           lcen, bcen - Galactic longitude and latitude of the center of the disk (deg)
           dist - distance in kpc
           radius - radius of the disk (deg)
        OUTPUT:
           (pixarea,extinction) - arrays of pixel-area in sq rad and extinction value
        HISTORY:
           2015-03-07 - Written - Bovy (IAS)
        """
        # Find all of the (l,b) of the pixels within radius of (lcen,bcen)
        lmin= round((lcen-radius-self._dl)/self._dl)*self._dl
        lmax= round((lcen+radius+self._dl)/self._dl)*self._dl
        bmin= round((bcen-radius-self._db)/self._db)*self._db
        bmax= round((bcen+radius+self._db)/self._db)*self._db
        ls= numpy.arange(lmin,lmax+self._dl,self._dl)
        bs= numpy.arange(bmin,bmax+self._db,self._db)
        ll,bb= numpy.meshgrid(ls,bs,indexing='ij')
        ll= ll.flatten()
        bb= bb.flatten()
        indx= cos_sphere_dist(numpy.sin((90.-bb)*_DEGTORAD),
                              numpy.cos((90.-bb)*_DEGTORAD),
                              numpy.sin(ll*_DEGTORAD),
                              numpy.cos(ll*_DEGTORAD),
                              numpy.sin((90.-bcen)*_DEGTORAD),
                              numpy.cos((90.-bcen)*_DEGTORAD),
                              numpy.sin(lcen*_DEGTORAD),
                              numpy.cos(lcen*_DEGTORAD)) \
                              >= numpy.cos(radius*_DEGTORAD)
        ll= ll[indx]
        bb= bb[indx]
        # Now get the extinctions for these pixels
        pixarea= self._dl*self._db*_DEGTORAD**2.+numpy.zeros(numpy.sum(indx))
        extinction= []
        for l,b in zip(ll,bb):
            extinction.append(self._evaluate(l,b,dist))
        extinction= numpy.array(extinction)
        return (pixarea,extinction)
               
    def dmax(self,l,b):
        """
        NAME:
           dmax
        PURPOSE:
           return the maximum distance for which there is Marshall et al. 
           (2006) data
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
        OUTPUT:
           maximum distance in kpc
        HISTORY:
           2013-12-19 - Started - Bovy (IAS)
        """
        tlbData= self.lbData(l,b,addBC=False)
        return tlbData['dist'][-1]

    def lbData(self,l,b,addBC=False):
        """
        NAME:
           lbData
        PURPOSE:
           return the Marshall et al. (2006) data corresponding to a given
           line of sight
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
           addBC= (False) if True, add boundary conditions (extinction is zero at zero distance; extinction is constant after last data point)
        OUTPUT:
        HISTORY:
           2013-12-13 - Written - Bovy (IAS)
        """
        #Find correct entry
        lbIndx= self._lbIndx(l,b)
        #Build output array
        out= numpy.recarray((self._marshalldata[lbIndx]['nb']+2*addBC,),
                            dtype=[('dist', 'f8'),
                                   ('e_dist', 'f8'),
                                   ('aks', 'f8'),
                                   ('e_aks','f8')])
        if addBC:
            #Add boundary conditions
            out[0]['dist']= 0.
            out[0]['e_dist']= 0.
            out[0]['aks']= 0.
            out[0]['e_aks']= 0.
            out[-1]['dist']= 30.
            out[-1]['e_dist']= 0.
            out[-1]['aks']= self._marshalldata[lbIndx]['ext%i' % (self._marshalldata[lbIndx]['nb'])]
            out[-1]['e_aks']=  self._marshalldata[lbIndx]['e_ext%i' % (self._marshalldata[lbIndx]['nb'])]
        for ii in range(self._marshalldata[lbIndx]['nb']):
            out[ii+addBC]['dist']= self._marshalldata[lbIndx]['r%i' % (ii+1)]
            out[ii+addBC]['e_dist']= self._marshalldata[lbIndx]['e_r%i' % (ii+1)]
            out[ii+addBC]['aks']= self._marshalldata[lbIndx]['ext%i' % (ii+1)]
            out[ii+addBC]['e_aks']= self._marshalldata[lbIndx]['e_ext%i' % (ii+1)]
        return out

    def plotData(self,l,b,*args,**kwargs):
        """
        NAME:
           plotData
        PURPOSE:
           plot the Marshall et al. (2006) extinction values 
           along a given line of sight as a function of 
           distance
        INPUT:
           l,b - Galactic longitude and latitude (degree)
           bovy_plot.plot args and kwargs
        OUTPUT:
           plot to output device
        HISTORY:
           2013-12-15 - Written - Bovy (IAS)
        """
        if not _BOVY_PLOT_LOADED:
            raise NotImplementedError("galpy.util.bovy_plot could not be loaded, so there is no plotting; might have to install galpy (http://github.com/jobovy/galpy) for plotting")
        #First get the data
        tdata= self.lbData(l,b)
        #Filter
        if self._filter is None:
            filterFac= 1./aebv('2MASS Ks',sf10=self._sf10)
        else:
            filterFac= 1./aebv('2MASS Ks',sf10=self._sf10)\
                *aebv(self._filter,sf10=self._sf10)
        #Plot
        out= bovy_plot.plot(tdata['dist'],tdata['aks']*filterFac,
                            *args,**kwargs)
        #uncertainties
        pyplot.errorbar(tdata['dist'],tdata['aks']*filterFac,
                        xerr=tdata['e_dist'],
                        yerr=tdata['e_aks']*filterFac,
                        ls='none',marker=None,color='k')
        return out

    def _lbIndx(self,l,b):
        """Return the index in the _marshalldata array corresponding to this (l,b)"""
        if l <= -100.125 or l >= 100.125 or b <= -10.125 or b >= 10.125:
            raise IndexError("Given (l,b) pair not within the region covered by the Marshall et al. (2006) dust map")
        lIndx= int(round((l+100.)/self._dl))
        bIndx= int(round((b+10.)/self._db))
        return lIndx*81+bIndx

    @classmethod
    def download(cls, test=False):
        marshall_folder_path = os.path.join(dust_dir, "marshall06")
        marshall_path = os.path.join(marshall_folder_path, "table1.dat.gz")
        marshall_readme_path = os.path.join(dust_dir, "marshall06", "ReadMe")
        if not os.path.exists(marshall_path[:-3]):
            if not os.path.exists(marshall_folder_path):
                os.mkdir(marshall_folder_path)
            _MARSHALL_URL= "https://cdsarc.cds.unistra.fr/ftp/J/A+A/453/635/table1.dat.gz"
            downloader(_MARSHALL_URL, marshall_path, cls.__name__, test=test)
            if not test:
                with open(marshall_path, "rb") as inf, open(os.path.join(marshall_folder_path, "table1.dat"), "w", encoding="utf8") as tof:
                    decom_str = gzip.decompress(inf.read()).decode("utf-8")
                    tof.write(decom_str)
                os.remove(marshall_path)
        if not os.path.exists(marshall_readme_path):
            _MARSHALL_README_URL= "https://cdsarc.cds.unistra.fr/ftp/J/A+A/453/635/ReadMe"
            downloader(_MARSHALL_README_URL, marshall_readme_path, f"{cls.__name__} (ReadMe)", test=test)
        return None



###############################################################################
#
#   Zero: model with zero extinction
#
###############################################################################
import numpy
from mwdust.DustMap3D import DustMap3D
_DEGTORAD= numpy.pi/180.
class Zero(DustMap3D):
    """model with zero extinction"""
    def __init__(self,filter=None,sf10=True):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the zero extinction model
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
        OUTPUT:
           Instance
        HISTORY:
           2015-03-07 - Written - Bovy (IAS)
        """
        DustMap3D.__init__(self,filter=filter)
        self._sf10= sf10
        return None

    def _evaluate(self,l,b,d):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the dust-map
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
           d- distance (kpc) can be array
        OUTPUT:
           extinction E(B-V)
        HISTORY:
           2015-03-07 - Written - Bovy (IAS)
        """
        return numpy.zeros_like(d)

    def dust_vals_disk(self,lcen,bcen,dist,radius):
        """
        NAME:
           dust_vals_disk
        PURPOSE:
           return the distribution of extinction within a small disk as samples
        INPUT:
           lcen, bcen - Galactic longitude and latitude of the center of the disk (deg)
           dist - distance in kpc
           radius - radius of the disk (deg)
        OUTPUT:
           (pixarea,extinction) - arrays of pixel-area in sq rad and extinction value
        HISTORY:
           2015-03-06 - Written - Bovy (IAS)
        """
        pixarea= (1.-numpy.cos(radius*_DEGTORAD))*2.*numpy.pi
        return (numpy.array([pixarea]),numpy.zeros((1,len(dist))))




###############################################################################
#
#   DustMap3D: top-level class for a 3D dust map; all other dust maps inherit
#              from this
#
###############################################################################
import numpy
try:
    from galpy.util import plot as bovy_plot
    _BOVY_PLOT_LOADED= True
except ImportError:
    _BOVY_PLOT_LOADED= False

class DustMap3D:
    """top-level class for a 3D dust map; all other dust maps inherit from this"""
    def __init__(self,filter=None):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the dust map
        INPUT:
           filter= filter to return the extinction in when called
        OUTPUT:
        HISTORY:
           2013-11-24 - Started - Bovy (IAS)
        """
        self._filter= filter
        self.download()  # download the map

    def __call__(self,*args,**kwargs):
        """
        NAME:
           __call__
        PURPOSE:
           evaluate the dust map
        INPUT:
           Either:
              (l,b,d) -  Galactic longitude, latitude (deg), and distance (kpc)
        OUTPUT:
        HISTORY:
           2013-11-24 - Started - Bovy (IAS)
        """
        if True: #(l,b,d)
            l,b,d= args
            if isinstance(d,(int,float,numpy.float32,numpy.float64)):
                d= numpy.array([d])
            try:
                return self._evaluate(l,b,d,**kwargs)
            except AttributeError:
                raise NotImplementedError("'_evaluate' for this DustMap3D not implemented yet")

    def plot(self,l,b,*args,**kwargs):
        """
        NAME:
           plot
        PURPOSE:
           plot the extinction along a given line of sight as a function of 
           distance
        INPUT:
           l,b - Galactic longitude and latitude (degree)
           range= distance range in kpc
           distmod= (False) if True, plot as a function of distance modulus (range is distmod range)
           bovy_plot.plot args and kwargs
        OUTPUT:
           plot to output device
        HISTORY:
           2013-12-11 - Written - Bovy (IAS)
        """
        if not _BOVY_PLOT_LOADED:
            raise NotImplementedError("galpy.util.bovy_plot could not be loaded, so there is no plotting; might have to install galpy (http://github.com/jobovy/galpy) for plotting")
        distmod= kwargs.pop('distmod',False)
        range= kwargs.pop('range',None)
        if range is None and distmod:
            range= [4.,19.]
        else:
            range= [0.,12.]
        nds= kwargs.get('nds',101)
        #First evaluate the dust map
        ds= numpy.linspace(range[0],range[1],nds)
        if distmod:
            adust= self(l,b,10.**(ds/5.-2.))
        else:
            adust= self(l,b,ds)
        #Add labels
        if distmod:
            kwargs['xlabel']= r'$\mathrm{Distance\ modulus}$'
        else:
            kwargs['xlabel']= r'$D\,(\mathrm{kpc})$'
        if not self._filter is None:
            kwargs['ylabel']= r'$A_{%s}\,(\mathrm{mag})$' % (self._filter.split(' ')[-1])
        else:
            kwargs['ylabel']= r'$E(B-V)\,(\mathrm{mag})$'
        return bovy_plot.plot(ds,adust,*args,**kwargs)

    @classmethod
    def download(cls, test=False):
        pass



import os, os.path
import numpy
import struct
from mwdust.util import fortranfile 
_DRIMMELDIR= os.path.join(os.path.dirname(os.path.realpath(__file__)),
                          'drimmeldata')
def readDrimmelAll():
    out= {}
    allFilenames= ['avgrid.dat','avori2.dat','rf_allsky.dat',
               'avdisk.dat','avloc.dat','avspir.dat',
               'avdloc.dat','avori.dat']
    for filename in allFilenames:
        f = fortranfile.FortranFile(os.path.join(_DRIMMELDIR,filename))
        nx, ny, nz= 151, 151, 51
        if 'avori2' in filename:
            nx, ny, nz= 101, 201, 51
        elif 'avori' in filename:
            nx, ny, nz= 76, 151, 51
        elif 'avloc' in filename:
            nx, ny, nz= 101, 201, 51
        elif 'avdloc' in filename:
            nx, ny, nz= 31, 31, 51
        if not 'rf_allsky' in filename:
            out[filename.split('.')[0]]=\
                f.readReals(prec='f').reshape((nz,ny,nx)).T
        else:
            #Need to do more work to read this file
            rec= f._read_exactly(4) #Read the header
            rec= f._read_exactly(4*393216)
            num = len(rec)//struct.calcsize('i')
            out_rf_pixnum= numpy.array(struct.unpack(f.ENDIAN+str(num)+'i',
                                                     rec),dtype='int')
            rec= f._read_exactly(4*393216)
            num = len(rec)//struct.calcsize('i')
            out_rf_comp= numpy.array(struct.unpack(f.ENDIAN+str(num)+'i',
                                                   rec),dtype='int')
            rec= f._read_exactly(4*393216)
            num = len(rec)//struct.calcsize('f')
            out_rf_glon= numpy.array(struct.unpack(f.ENDIAN+str(num)+'f',
                                                   rec),dtype=numpy.float32)
            rec= f._read_exactly(4*393216)
            num = len(rec)//struct.calcsize('f')
            out_rf_glat= numpy.array(struct.unpack(f.ENDIAN+str(num)+'f',
                                                   rec),dtype=numpy.float32)
            rec= f._read_exactly(4*393216)
            num = len(rec)//struct.calcsize('f')
            out_rf= numpy.array(struct.unpack(f.ENDIAN+str(num)+'f',
                                              rec),dtype=numpy.float32)
            out['rf_pixnum']= out_rf_pixnum
            out['rf_comp']= out_rf_comp
            out['rf_glon']= out_rf_glon
            out['rf_glat']= out_rf_glat
            out['rf']= out_rf
        f.close()
    return out



import sys
import distutils.sysconfig as sysconfig
import ctypes
import ctypes.util
from numpy.ctypeslib import ndpointer
import numpy as np
from pathlib import Path

# healpy number to represent bad numbers
UNSEEN = -1.6375e+30

# Find and load the library
_lib = None
_libname = ctypes.util.find_library("healpix_c")
_ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
if _libname:
    _lib = ctypes.CDLL(_libname)
if _lib is None:
    # Add top-level mwdust repository directory for pip install (-e) .,
    # just becomes site-packages for regular install
    paths = sys.path
    paths.append(str(Path(__file__).parent.parent.parent.absolute()))
    for path in [Path(p) for p in paths]:
        if not path.is_dir():
            continue
        try:
            _lib = ctypes.CDLL(str(path / f"healpix_c{_ext_suffix}"))
        except OSError:
            _lib = None
        else:
            break
if _lib is None:
    raise IOError("healpix/C module not found")


# useful utilities: http://graphics.stanford.edu/~seander/bithacks.html#DetermineIfPowerOf2
def check_nside(nside, nest=False):
    """
    Utility function
    
    Check if healpix map with nside is valid in general
    """
    # nside can only be a power of 2 for nest, but generally less than 2**29
    nside_arr = np.array(nside).astype(np.int64)
    is_ok = True
    if np.all(np.logical_and(np.logical_and(nside == nside_arr, np.all(np.less(0, nside))), nside_arr <= 2**29)):
        if nest:
            is_ok = (nside_arr & (nside_arr - 1)) == 0
    else:
        is_ok = False
    if not is_ok:
        raise ValueError(f"{nside} is not valid")


def check_npix(npix):
    """
    Utility function
    
    Check if total pixel number of a healpix map are valid in general
    """
    # check if npix is a valid value for healpix map size
    nside = np.sqrt(np.asarray(npix) / 12.0)
    if nside != np.floor(nside):
        raise ValueError(f"{npix} is not a valid value for healpix map size")


def check_ipix_nside(ipix, nside):
    """
    Utility function
    
    Check if pixel number(s) are valid in a healpix map with nside
    """
    # check if all ipix are valid for a healpix map size
    if not np.all(ipix <= nside2npix(nside)):
        raise ValueError(f"Not all ipix are valid for such healpix map size")


def npix2nside(npix):
    """
    Utility function

    Give the nside parameter for the given number of pixels.
    """
    check_npix(npix)
    return int(np.sqrt(npix / 12.0))


def lonlat2thetaphi(lon, lat):
    """
    NAME:
        lonlat2thetaphi
    PURPOSE:
        Angular coordinates convsrion: longitude and latitude (deg) to colatitude/longitude (rad)
    INPUT:
        lon, lat - longitude and latitude (deg)
    OUTPUT:
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    return np.pi / 2.0 - np.radians(lat), np.radians(lon)


def thetaphi2lonlat(theta, phi):
    """
    NAME:
        thetaphi2lonlat
    PURPOSE:
        Angular coordinates convsrion: colatitude/longitude (rad) to longitude and latitude (deg)
    INPUT:
        theta, phi - colatitude/longitude (rad)
    OUTPUT:
        lon, lat - longitude and latitude (deg)
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    return np.degrees(phi), 90.0 - np.degrees(theta)


def ang2pix(nside, theta, phi, nest=False, lonlat=False):
    """
    NAME:
        ang2pix
    PURPOSE:
        Angular coordinates to healpix map pixel number
    INPUT:
        nside - a integer of healpix map nside
        theta - colatitude
        phi - longitude
        nest - is using NEST?
        lonlat - input in longitude and latitude (deg)?
    OUTPUT:
        ipix - pixel numbers
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    check_nside(nside)
    theta, phi = np.atleast_1d(np.asarray(theta)), np.atleast_1d(np.asarray(phi))
    nstars = len(theta)
    if lonlat:
        theta, phi = lonlat2thetaphi(theta, phi)
    if nest is False:
        raise NotImplementedError("RING scheme is not avaliable for now")
        # ang2pix_c = _lib.ang2pix_ring
    else:
        ang2pix_c = _lib.ang2pix_nest
    ndarrayFlags = ("C_CONTIGUOUS", "WRITEABLE")
    
    ang2pix_c.argtypes = [
        ctypes.c_long,
        ndpointer(dtype=np.float64, flags=ndarrayFlags),
        ndpointer(dtype=np.float64, flags=ndarrayFlags),
        ctypes.c_long,
    ]
    ang2pix_c.restype = ctypes.POINTER(ctypes.c_longlong)

    # Array requirements, first store old order
    f_cont = [theta.flags["F_CONTIGUOUS"], phi.flags["F_CONTIGUOUS"]]
    theta = np.require(theta, dtype=np.float64, requirements=["C", "W"])
    phi = np.require(phi, dtype=np.float64, requirements=["C", "W"])
    res = ang2pix_c(
        ctypes.c_long(nside),
        theta.astype(np.float64, order="C", copy=False),
        phi.astype(np.float64, order="C", copy=False),
        ctypes.c_long(nstars),
    )
    result = np.fromiter(res, dtype=np.int32, count=nstars)

    # Reset input arrays
    if f_cont[0]:
        theta = np.asfortranarray(theta)
    if f_cont[1]:
        phi = np.asfortranarray(phi)

    return result


def ang2vec(theta, phi, lonlat=False):
    """
    NAME:
        ang2vec
    PURPOSE:
        Angular coordinates to unit 3-vector direction
    INPUT:
        theta - colatitude
        phi - longitude 
        lonlat - input in longitude and latitude (deg)?
    OUTPUT:
        x, y, z - unit 3-vector direction
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    theta, phi = np.atleast_1d(np.asarray(theta)), np.atleast_1d(np.asarray(phi))
    nstars = len(theta)
    if lonlat:
        theta, phi = lonlat2thetaphi(theta, phi)

    ang2vec_c = _lib.ang2vec
    ndarrayFlags = ("C_CONTIGUOUS", "WRITEABLE")
    
    ang2vec_c.argtypes = [
        ndpointer(dtype=np.float64, flags=ndarrayFlags),
        ndpointer(dtype=np.float64, flags=ndarrayFlags),
        ctypes.c_long,
    ]
    ang2vec_c.restype = ctypes.POINTER(ctypes.c_double)

    # Array requirements, first store old order
    f_cont = [theta.flags["F_CONTIGUOUS"], phi.flags["F_CONTIGUOUS"]]
    theta = np.require(theta, dtype=np.float64, requirements=["C", "W"])
    phi = np.require(phi, dtype=np.float64, requirements=["C", "W"])
    res = ang2vec_c(
        theta.astype(np.float64, order="C", copy=False),
        phi.astype(np.float64, order="C", copy=False),
        ctypes.c_long(nstars),
    )
    result = np.fromiter(res, dtype=np.float64, count=nstars*3)
    result = result.reshape(nstars, 3)

    # Reset input arrays
    if f_cont[0]:
        theta = np.asfortranarray(theta)
    if f_cont[1]:
        phi = np.asfortranarray(phi)

    return result

def pix2vec(nside, ipix, nest=False):
    """
    NAME:
        pix2vec
    PURPOSE:
        Pixel number to unit 3-vector direction
    INPUT:
        nside - a integer of healpix map nside
        ipix - a (list) of integer of pixel number
        nest - is using NEST?
    OUTPUT:
        x, y, z - unit 3-vector direction
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    ipix = np.asarray(ipix)
    check_nside(nside, nest=nest)
    check_ipix_nside(ipix, nside)
    ipix = np.atleast_1d(ipix)
    npix = len(ipix)
    if nest:
        pix2vec_c = _lib.pix2vec_nest
    else:
        raise NotImplementedError("RING scheme is not avaliable for now")
        # _lib.pix2ang_ring
    ndarrayFlags = ("C_CONTIGUOUS", "WRITEABLE")
    # (int nside, int npix, long long int *ipix)
    pix2vec_c.argtypes = [
        ctypes.c_long,
        ctypes.c_long,
        ndpointer(dtype=np.int64, flags=ndarrayFlags),
    ]
    pix2vec_c.restype = ctypes.POINTER(ctypes.c_double)

    # Array requirements, first store old order
    f_cont = [ipix.flags["F_CONTIGUOUS"]]
    ipix = np.require(ipix, dtype=np.int64, requirements=["C", "W"])
    res = pix2vec_c(
        nside,
        npix,
        ipix.astype(np.int64, order="C", copy=False),
    )
    result = np.fromiter(res, dtype=np.float64, count=npix*3)
    result = result.reshape(npix, 3)

    # Reset input arrays
    if f_cont[0]:
        ipix = np.asfortranarray(ipix)

    x, y, z = result[:, 0], result[:, 1], result[:, 2]
    return x, y, z


def pix2ang(nside, ipix, nest=False, lonlat=False):
    """
    NAME:
        pix2ang
    PURPOSE:
        Pixel number to angular coordinates
    INPUT:
        nside - a integer of healpix map nside
        ipix - a (list) of integer of pixel number
        nest - is using NEST?
        lonlat - output in longitude and latitude (deg)?
    OUTPUT:
        theta, phi - angular coordinates
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    ipix = np.asarray(ipix)
    check_nside(nside, nest=nest)
    check_ipix_nside(ipix, nside)
    ipix = np.atleast_1d(ipix)
    npix = len(ipix)
    if nest:
        pix2ang_c = _lib.pix2ang_nest
    else:
        raise NotImplementedError("RING scheme is not avaliable for now")
        # _lib.pix2ang_ring
    ndarrayFlags = ("C_CONTIGUOUS", "WRITEABLE")
    # (int nside, int npix, long long int *ipix)
    pix2ang_c.argtypes = [
        ctypes.c_long,
        ctypes.c_long,
        ndpointer(dtype=np.int64, flags=ndarrayFlags),
    ]
    pix2ang_c.restype = ctypes.POINTER(ctypes.c_double)

    # Array requirements, first store old order
    f_cont = [ipix.flags["F_CONTIGUOUS"]]
    ipix = np.require(ipix, dtype=np.int64, requirements=["C", "W"])
    res = pix2ang_c(
        nside,
        npix,
        ipix.astype(np.int64, order="C", copy=False),
    )
    result = np.fromiter(res, dtype=np.float64, count=npix*2)
    result = result.reshape(npix, 2)

    # Reset input arrays
    if f_cont[0]:
        ipix = np.asfortranarray(ipix)

    theta, phi = result[:, 0], result[:, 1]

    if lonlat:
        return thetaphi2lonlat(theta, phi)
    else:
        return theta, phi


def nside2pixarea(nside):
    """
    NAME:
        nside2pixarea
    PURPOSE:
        Get area of a pixel for a healpix map with nside
    INPUT:
        nside - a integer of healpix map nside
    OUTPUT:
        pixarea - a float of healpix map pixel area
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    return np.pi / (3 * nside ** 2)


def nside2npix(nside):
    """
    NAME:
        nside2npix
    PURPOSE:
        Get how many pixel for a healpix map with nside
    INPUT:
        nside - a integer of healpix map nside
    OUTPUT:
        npix - a integer of number of healpix map pixel
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    return 12 * nside * nside


def ud_grade(map_in, nside_plot, pess=False, order_in="RING", order_out=None, power=None, dtype=None):
    """
    NAME:
        dust_vals_disk
    PURPOSE:
        Upgrade or degrade healpix map
    INPUT:
        map_in - healpix map to be upgraded or degraded
        nside_plot - nside of the output map wanted
    OUTPUT:
        map_out - arrays of upgraded or degraded map(s)
    HISTORY:
        2023-03-01 - Written - Henry Leung (Toronto)
    """
    # check if arguements are implemented
    if order_in != "NEST" or (order_out is not None and order_out != "NEST"):
        raise NotImplementedError("order_in and order_out for RING scheme is not implemented for now")
    if power is not None:
        raise NotImplementedError("power is not implemented for now")
    if pess is not False:
        raise NotImplementedError("pess=True is not implemented for now")
    if dtype is not False:
        raise NotImplementedError("dtype is implemented for now, output map always has the same dtype as input map")
    check_nside(nside_plot, nest=order_in != "RING")
    map_in = np.asarray(map_in)

    num_of_map = len(np.atleast_2d(map_in))
    if num_of_map != 1:
        raise ValueError("This function only support one map at each time")
    
    nside_in = npix2nside(len(map_in))
    npix_in = nside2npix(nside_in)
    npix_out = nside2npix(nside_in)

    if nside_plot > nside_in: # upgrade
        rat2 = npix_out // npix_in
        fact = np.ones(rat2, dtype=map_in.dtype)
        map_out = np.outer(map_in, fact).reshape(npix_out)
    elif nside_plot < nside_in: # degrade
        rat2 = npix_in // npix_out
        mr = map_in.reshape(npix_out, rat2)
        goods = ~(np.isclose(mr, UNSEEN) | (~np.isfinite(mr)) | (~np.isnan(mr)))
        map_out = np.sum(mr * goods, axis=1)
        nhit = goods.sum(axis=1)
        map_out[nhit != 0] = map_out[nhit != 0] / nhit[nhit != 0]
        try:
            map_out[nhit == 0] = UNSEEN
        except OverflowError:
            pass
    else:
        map_out = map_in
        
    return map_out


###############################################################################
#
#   Sale14: extinction model from Sale et al. 2014 2014MNRAS.443.2907S
#
###############################################################################
import os, os.path
import sys
import tarfile
import shutil
import numpy
from scipy import interpolate
from astropy.io import ascii
from mwdust.util.extCurves import aebv
from mwdust.util.tools import cos_sphere_dist
from mwdust.util.download import downloader, dust_dir
from mwdust.DustMap3D import DustMap3D

_DEGTORAD= numpy.pi/180.
_saledir= os.path.join(dust_dir,'sale14')
_ERASESTR= "                                                                                "
class Sale14(DustMap3D):
    """extinction model from Sale et al. 2014 2014MNRAS.443.2907S"""
    def __init__(self,filter=None,sf10=True):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the Sale14 dust map
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
        OUTPUT:
           object
        HISTORY:
           2015-03-08 - Started - Bovy (IAS)
        """
        DustMap3D.__init__(self,filter=filter)
        self._sf10= sf10
        #Read the maps
        sys.stdout.write('\r'+"Reading Sale et al. (2014) data file ...\r")
        sys.stdout.flush()
        self._saledata= ascii.read(os.path.join(_saledir,
                                                     'Amap.dat'),
                                        readme=os.path.join(_saledir,
                                                            'ReadMe'),
                                        guess=False, format='cds',
                                        fill_values=[('', '-999')])
        sys.stdout.write('\r'+_ERASESTR+'\r')
        sys.stdout.flush()
        # Some summaries
        self._dl= self._saledata['lmax']-self._saledata['lmin']
        self._db= self._saledata['b_max']-self._saledata['b_min']
        self._lmin= numpy.amin(self._saledata['lmin'])
        self._lmax= numpy.amax(self._saledata['lmax'])
        self._bmin= numpy.amin(self._saledata['b_min'])
        self._bmax= numpy.amax(self._saledata['b_max'])
        self._ndistbin= 150
        self._ds= numpy.linspace(0.05,14.95,self._ndistbin)
        # For dust_vals
        self._sintheta= numpy.sin((90.-self._saledata['GLAT'])*_DEGTORAD)
        self._costheta= numpy.cos((90.-self._saledata['GLAT'])*_DEGTORAD)
        self._sinphi= numpy.sin(self._saledata['GLON']*_DEGTORAD)
        self._cosphi= numpy.cos(self._saledata['GLON']*_DEGTORAD)
        self._intps= numpy.zeros(len(self._saledata),dtype='object') #array to cache interpolated extinctions
        return None

    def _evaluate(self,l,b,d,_lbIndx=None):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the dust-map
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
           d- distance (kpc) can be array
        OUTPUT:
           extinction
        HISTORY:
           2015-03-08 - Started - Bovy (IAS)
        """
        if isinstance(l,numpy.ndarray) or isinstance(b,numpy.ndarray):
            raise NotImplementedError("array input for l and b for Sale et al. (2014) dust map not implemented")
        if _lbIndx is None: lbIndx= self._lbIndx(l,b)
        else: lbIndx= _lbIndx
        if self._intps[lbIndx] != 0:
            out= self._intps[lbIndx](d)
        else:
            tlbData= self.lbData(l,b)
            interpData=\
                interpolate.InterpolatedUnivariateSpline(self._ds,
                                                         tlbData['a0'],
                                                         k=1)
            out= interpData(d)
            self._intps[lbIndx]= interpData
        if self._filter is None: # Sale et al. say A0/Aks = 11
            return out/11./aebv('2MASS Ks',sf10=self._sf10)
        else: # if sf10, first put ebv on SFD scale
            return out/11./aebv('2MASS Ks',sf10=self._sf10)\
                *aebv(self._filter,sf10=self._sf10)

    def dust_vals_disk(self,lcen,bcen,dist,radius):
        """
        NAME:
           dust_vals_disk
        PURPOSE:
           return the distribution of extinction within a small disk as samples
        INPUT:
           lcen, bcen - Galactic longitude and latitude of the center of the disk (deg)
           dist - distance in kpc
           radius - radius of the disk (deg)
        OUTPUT:
           (pixarea,extinction) - arrays of pixel-area in sq rad and extinction value
        HISTORY:
           2015-03-07 - Written - Bovy (IAS)
        """
        # Find all of the (l,b) of the pixels within radius of (lcen,bcen)
        indx= cos_sphere_dist(self._sintheta,self._costheta,
                              self._sinphi,self._cosphi,
                              numpy.sin((90.-bcen)*_DEGTORAD),
                              numpy.cos((90.-bcen)*_DEGTORAD),
                              numpy.sin(lcen*_DEGTORAD),
                              numpy.cos(lcen*_DEGTORAD)) \
                              >= numpy.cos(radius*_DEGTORAD)
        ll= self._saledata['GLON'][indx]
        bb= self._saledata['GLAT'][indx]
        # Now get the extinctions for these pixels
        pixarea= []
        extinction= []
        for l,b in zip(ll,bb):
            lbIndx= self._lbIndx(l,b)
            extinction.append(self._evaluate(l,b,dist,_lbIndx=lbIndx))
            pixarea.append(self._dl[lbIndx]*self._db[lbIndx]*_DEGTORAD**2.)
        pixarea= numpy.array(pixarea)
        extinction= numpy.array(extinction)
        return (pixarea,extinction)
               
    def dmax(self,l,b):
        """
        NAME:
           dmax
        PURPOSE:
           return the maximum distance for which to trust the Sale et al. (2014) data
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
        OUTPUT:
           maximum distance in kpc
        HISTORY:
           2015-03-08 - Written - Bovy (IAS)
        """
        lbIndx= self._lbIndx(l,b)
        return self._saledata['trust'][lbIndx]/1000.

    def lbData(self,l,b):
        """
        NAME:
           lbData
        PURPOSE:
           return the Sale et al. (2014) data corresponding to a given
           line of sight
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
        OUTPUT:
        HISTORY:
           2015-03-08 - Written - Bovy (IAS)
        """
        #Find correct entry
        lbIndx= self._lbIndx(l,b)
        #Build output array
        out= numpy.recarray((self._ndistbin,),
                            dtype=[('a0', 'f8'),
                                   ('e_a0','f8')])
        for ii in range(self._ndistbin):
            out[ii]['a0']= self._saledata[lbIndx]['meanA%i' % (ii+1)]
            out[ii]['e_a0']= self._saledata[lbIndx]['meanA%i' % (ii+1)]
        return out

    def _lbIndx(self,l,b):
        """Return the index in the _saledata array corresponding to this (l,b)"""
        if l <= self._lmin or l >= self._lmax \
                or b <= self._bmin or b >= self._bmax:
            raise IndexError("Given (l,b) pair not within the region covered by the Sale et al. (2014) dust map")
        return numpy.argmin((l-self._saledata['GLON'])**2./self._dl**2.\
                                +(b-self._saledata['GLAT'])**2./self._db**2.)

    @classmethod
    def download(cls, test=False):
        sale_folder_path = os.path.join(dust_dir, "sale14")
        sale_path = os.path.join(sale_folder_path, "Amap.tar.gz")
        if not os.path.exists(sale_path[:-6] + "dat"):
            if not os.path.exists(sale_folder_path):
                os.mkdir(sale_folder_path)
            _SALE_URL= "http://www.iphas.org/data/extinction/Amap.tar.gz"
            downloader(_SALE_URL, sale_path, cls.__name__, test=test)
            if not test:
                sale_file = tarfile.open(sale_path)
                sale_file.extractall(sale_folder_path)
                sale_file.close()
                os.remove(sale_path)
                # Fix one line in the dust map
                with open(os.path.join(sale_folder_path, "tmp.dat"), "w") as fout:
                    with open(os.path.join(sale_folder_path, "Amap.dat"), "r") as fin:
                        for line in fin:
                            if "15960.40000" in line: # bad line
                                newline= ''
                                for ii,word in enumerate(line.split(' ')):
                                    if ii > 0: newline+= ' '
                                    if ii > 6 and len(word) > 9:
                                        word= "747.91400"
                                    newline+= word
                                fout.write(newline+'\n')
                            else:
                                fout.write(line)
                shutil.move(os.path.join(sale_folder_path, "tmp.dat"), os.path.join(sale_folder_path, "Amap.dat"))
        return None



# combine_dustmaps.py: make a composite extinction map (Marshall,Green,Drimmel)
# Using legacy code and only runs in python 2
# Creates a composite map. Run from command line:
# python combine_dustmaps19.py combine19.sav (runs the combine_dustmap function)
# python combine_dustmaps19.py combine19.sav save (runs the store_h5 function)
# python combine_dustmaps19.py combine19.sav fixnans (runs the fix_nans function)
# The first call takes a few hours, can be tested via _DRYISHRUN = False first
###############################################################################
import sys
import os, os.path
import pickle
import numpy
import h5py
import healpy
import mwdust
from mwdust.util.download import dust_dir

from galpy.util import save_pickles


def distmod2dist(distmod):
    """distance modulus to distance in kpc"""
    return 10.**(distmod/5.-2.)

_greendir= os.path.join(dust_dir,'green19')
#_GREEN15DISTMODS= numpy.linspace(4.,19.,31) # distance spacing for green15 and green17
_GREEN19DISTMODS= numpy.linspace(4,18.875,120) # distance spacing for green19
_GREEN19DISTS= distmod2dist(_GREEN19DISTMODS)
_DEGTORAD= numpy.pi/180.
_ERASESTR= "                                                                                "
_DRYISHRUN= False
def combine_dustmap(picklename):
    if os.path.exists(picklename): return None
    ndists= len(_GREEN19DISTS)
    # First fill in with NSIDE = 512 for Marshall
    marshallmap= mwdust.Marshall06()
    nside_mar= 512
    mar_pix= numpy.arange(healpy.pixelfunc.nside2npix(nside_mar))
    mar_val= numpy.zeros((len(mar_pix),ndists))+healpy.UNSEEN
    theta, phi= \
        healpy.pixelfunc.pix2ang(nside_mar,mar_pix,nest=True)
    bb= (numpy.pi/2.-theta)/numpy.pi*180.
    ll= phi/numpy.pi*180.
    subIndx= (numpy.fabs(bb) < 10.125)\
        *((ll < 100.125)+(ll > 259.875))
    mar_pix= mar_pix[subIndx]
    mar_val= mar_val[subIndx]
    ll= ll[subIndx]
    ll[ll > 180.]-= 360.
    bb= bb[subIndx]
    for pp,dpix in enumerate(mar_pix):
        sys.stdout.write('\r'+"Working on pixel %i, %i remaining ...\r" % (pp+1,len(mar_pix)-pp))
        sys.stdout.flush()
        if _DRYISHRUN and pp > 100: break
        mar_val[pp]= marshallmap(ll[pp],bb[pp],_GREEN19DISTS)
    sys.stdout.write('\r'+_ERASESTR+'\r')
    sys.stdout.flush()
    # Now load Green19 and remove those pixels that fall within the Marshall map
    with h5py.File(os.path.join(_greendir,'bayestar2019.h5'),'r') as greendata:
        pix_info= greendata['/pixel_info'][:]
        best_fit= greendata['/best_fit'][:]
    # Check whether any of these fall within the Marshall map
    theta, phi= healpy.pixelfunc.pix2ang(pix_info['nside'].astype('int32'),
                                         pix_info['healpix_index'].astype('int64'),
                                         nest=True)
    inMar= ((phi < 100.125*_DEGTORAD)+(phi > 259.875*_DEGTORAD))\
        *(numpy.fabs(numpy.pi/2.-theta) < 10.125*_DEGTORAD)
    best_fit[inMar]= healpy.UNSEEN
    nside_min= numpy.min(pix_info['nside'])
    nside_max= numpy.max(pix_info['nside'])
    # Fill in remaining gaps with Drimmel at NSIDE=256
    pix_drim= []
    pix_drim_nside= []
    val_drim= []
    drimmelmap= mwdust.Drimmel03()   
    for nside_drim in 2**numpy.arange(8,int(numpy.log2(nside_max))+1,1):
        tpix= numpy.arange(healpy.pixelfunc.nside2npix(nside_drim))
        rmIndx= numpy.zeros(len(tpix),dtype='bool')
        # Remove pixels that already have values at this or a higher level
        for nside in 2**numpy.arange(8,
                                     int(numpy.log2(nside_max))+1,1):
            mult_factor = (nside/nside_drim)**2
            tgpix= pix_info['healpix_index'][pix_info['nside'] == nside]
            for offset in numpy.arange(mult_factor):
                rmIndx[numpy.in1d(tpix*mult_factor+offset,tgpix,
                                  assume_unique=True)]= True
        # Remove pixels that already have values at a lower level
        for nside in 2**numpy.arange(int(numpy.log2(nside_min)),
                                         int(numpy.log2(nside_drim)),1):
            mult_factor = (nside_drim/nside)**2
            # in Green 19
            tgpix= pix_info['healpix_index'][pix_info['nside'] == nside]
            rmIndx[numpy.in1d(tpix//mult_factor,tgpix,assume_unique=False)]= True
            # In the current Drimmel
            tdpix= numpy.array(pix_drim)[numpy.array(pix_drim_nside) == nside]
            rmIndx[numpy.in1d(tpix//mult_factor,tdpix,assume_unique=False)]= True
        # Also remove pixels that lie within the Marshall area
        theta, phi= healpy.pixelfunc.pix2ang(nside_drim,tpix,nest=True)
        inMar= ((phi < 100.125*_DEGTORAD)+(phi > 259.875*_DEGTORAD))\
            *(numpy.fabs(numpy.pi/2.-theta) < 10.125*_DEGTORAD)
        rmIndx[inMar]= True
        tpix= tpix[True-rmIndx]
        pix_drim.extend(tpix)
        pix_drim_nside.extend(nside_drim*numpy.ones(len(tpix)))
        ll= phi[True-rmIndx]/_DEGTORAD
        bb= (numpy.pi/2.-theta[True-rmIndx])/_DEGTORAD
        for pp in range(len(tpix)):
            sys.stdout.write('\r'+"Working on level %i, pixel %i, %i remaining ...\r" % (nside_drim,pp+1,len(tpix)-pp))
            sys.stdout.flush()
            val_drim.append(drimmelmap(ll[pp],bb[pp],_GREEN19DISTS))
            if _DRYISHRUN and pp > 1000: break
    sys.stdout.write('\r'+_ERASESTR+'\r')
    sys.stdout.flush()
    # Save
    g19Indx= best_fit[:,0] != healpy.UNSEEN
    save_pickles(picklename,mar_pix,mar_val,
                 pix_info['nside'][g19Indx],pix_info['healpix_index'][g19Indx],
                 best_fit[g19Indx],
                 pix_drim,pix_drim_nside,val_drim)
    return None

def store_h5(picklename):
    # Restore pickle
    if not os.path.exists(picklename):
        print("file %s does not exist!" % picklename)
        return None
    with open(picklename,'rb') as picklefile:
        mar_pix= pickle.load(picklefile)
        mar_val= pickle.load(picklefile)
        pix_info_nside= pickle.load(picklefile)
        pix_info_healpix= pickle.load(picklefile)
        best_fit= pickle.load(picklefile)
        pix_drim= pickle.load(picklefile)
        pix_drim_nside= pickle.load(picklefile)
        val_drim= pickle.load(picklefile)
    # Combine
    nout= len(mar_pix)+len(pix_info_nside)+len(val_drim)
    out_nside= numpy.empty(nout,dtype='uint32')
    out_healpix= numpy.empty(nout,dtype='uint64')
    out_bf= numpy.empty((nout,len(_GREEN19DISTS)),dtype='float64')
    # Load Marshall
    out_nside[:len(mar_pix)]= 512
    out_healpix[:len(mar_pix)]= mar_pix
    out_bf[:len(mar_pix)]= mar_val
    # Load Green 19
    out_nside[len(mar_pix):len(mar_pix)+len(pix_info_nside)]= pix_info_nside
    out_healpix[len(mar_pix):len(mar_pix)+len(pix_info_nside)]= pix_info_healpix
    out_bf[len(mar_pix):len(mar_pix)+len(pix_info_nside)]= best_fit
    # Load Drimmel
    if _DRYISHRUN:
        out_nside[len(mar_pix)+len(pix_info_nside):]= pix_drim_nside[:len(val_drim)]
        out_healpix[len(mar_pix)+len(pix_info_nside):]= pix_drim[:len(val_drim)]
    else:
        out_nside[len(mar_pix)+len(pix_info_nside):]= pix_drim_nside
        out_healpix[len(mar_pix)+len(pix_info_nside):]= pix_drim
    out_bf[len(mar_pix)+len(pix_info_nside):]= val_drim
    # Save to h5 file
    outfile= h5py.File(picklename.replace('.sav','.h5'),"w")
    pixinfo= numpy.recarray(len(out_nside),
                            dtype=[('nside','uint32'),
                                   ('healpix_index','uint64')])
    pixinfo['nside']= out_nside
    pixinfo['healpix_index']= out_healpix
    outfile.create_dataset("pixel_info", data=pixinfo)
    outfile.create_dataset("best_fit",data=out_bf)
    outfile.close()
    return None    

def fix_nans(picklename):
    """Marshall has a few NaNs, replace these with Drimmel"""
    with h5py.File(picklename.replace('.sav','.h5'),'r') as combdata:
        pix_info= combdata['/pixel_info'][:]
        best_fit= combdata['/best_fit'][:]
    nanIndx= numpy.isnan(best_fit[:,0])
    print("Found %i NaNs ..." % numpy.sum(nanIndx))
    theta, phi= healpy.pixelfunc.pix2ang(pix_info['nside'].astype('int32'),
                                         pix_info['healpix_index'].astype('int64'),
                                         nest=True)
    bb= (numpy.pi/2.-theta)/_DEGTORAD
    ll= phi/_DEGTORAD
    indices= numpy.arange(len(pix_info['nside']))[nanIndx]
    drimmelmap= mwdust.Drimmel03()   
    for ii in range(numpy.sum(nanIndx)):
        best_fit[indices[ii]]= drimmelmap(ll[ii],bb[ii],_GREEN19DISTS)
    # Now save
    nanIndx= numpy.isnan(best_fit[:,0])
    print("Found %i NaNs ..." % numpy.sum(nanIndx))
    # Save to h5 file
    outfile= h5py.File(picklename.replace('.sav','.h5'),"w")
    outfile.create_dataset("pixel_info", data=pix_info)
    outfile.create_dataset("best_fit",data=best_fit)
    outfile.close()
    return None

if __name__ == '__main__':
    if len(sys.argv) > 2 and sys.argv[2] == 'fixnans':
        fix_nans(sys.argv[1])
    elif len(sys.argv) > 2:
        store_h5(sys.argv[1])
    else:
        print(sys.argv[1])        
        combine_dustmap(sys.argv[1])



import numpy
from numpy.random import default_rng
rng= default_rng()

def test_avebv():
    # Simple test of the extinction curve, using SFD as an example
    # Test that A_V/E(B-V) = 3.1 for the CTIO V filter
    from mwdust import SFD
    sfd_av= SFD(filter='CTIO V')
    sfd_ebv= SFD()
    glons= rng.uniform(0.,360.,size=20)
    glats= rng.uniform(-90.,90.,size=20)
    dists= rng.uniform(0.1,10.,size=5)
    for ll,bb,dist in zip(glons,glats,dists):
        assert numpy.fabs(sfd_av(ll,bb,dist)/sfd_ebv(ll,bb,dist)/0.86-3.1) < 0.02, \
            'SFD extinction curve for CTIO V does not agree with known values for A_V/E(B-V) = 3.1'
    return None



from mwdust.SFD import SFD
from mwdust.Marshall06 import Marshall06
from mwdust.Drimmel03 import Drimmel03
from mwdust.Sale14 import Sale14
from mwdust.Green15 import Green15
from mwdust.Green17 import Green17
from mwdust.Green19 import Green19
from mwdust.Combined15 import Combined15
from mwdust.Combined19 import Combined19
from mwdust.Zero import Zero

__version__ = "1.5.post0"


def download_all(test=False):
    SFD.download(test=test)
    Marshall06.download(test=test)
    Drimmel03.download(test=test)
    Sale14.download(test=test)
    Green15.download(test=test)
    Green17.download(test=test)
    Green19.download(test=test)
    Combined15.download(test=test)
    Combined19.download(test=test)



###############################################################################
#
#   SFD: Schlegel, Finkbeiner, & Davis (1998) dust map (2D)
#
###############################################################################
import os
import numpy
from mwdust.util.download import downloader
from mwdust.util.read_SFD import read_SFD_EBV
from mwdust.util.extCurves import aebv
from mwdust.util.download import downloader, dust_dir
from mwdust.DustMap3D import DustMap3D

class SFD(DustMap3D):
    """Schlegel, Finkbeiner, & Davis (1998) dust map (2D)"""
    def __init__(self,filter=None,sf10=True,interp=True,noloop=False):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the SFD dust map
        INPUT:
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
           filter= filter to return the extinction in
           interp= (True) if True, interpolate using the nearest pixels
           noloop= (False) if True, don't loop through the glons
        OUTPUT:
           object
        HISTORY:
           2013-11-24 - Started - Bovy (IAS)
        """
        DustMap3D.__init__(self,filter=filter)
        self._sf10= sf10
        self._interp= interp
        self._noloop= noloop
        return None

    def _evaluate(self,l,b,d):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the dust-map
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
           d- distance (kpc)
        OUTPUT:
           extinction
        HISTORY:
           2013-11-24 - Started - Bovy (IAS)
        """
        tebv= read_SFD_EBV(l,b,interp=self._interp,
                           noloop=self._noloop,verbose=False)
        if self._filter is None:
            return tebv*numpy.ones_like(d)
        else:
            return tebv*aebv(self._filter,sf10=self._sf10)*numpy.ones_like(d)

    @classmethod
    def download(cls, test=False):
        sfd_ngp_path = os.path.join(dust_dir, "maps", "SFD_dust_4096_ngp.fits")
        if not os.path.exists(sfd_ngp_path):
            if not os.path.exists(os.path.join(dust_dir, "maps")):
                os.mkdir(os.path.join(dust_dir, "maps"))
            _SFD_URL_NGP= "https://svn.sdss.org/public/data/sdss/catalogs/dust/trunk/maps/SFD_dust_4096_ngp.fits"
            downloader(_SFD_URL_NGP, sfd_ngp_path, "SFD (NGP)", test=test)
        sfd_sgp_path = os.path.join(dust_dir, "maps", "SFD_dust_4096_sgp.fits")
        if not os.path.exists(sfd_sgp_path):
            _SFD_URL_SGP= "https://svn.sdss.org/public/data/sdss/catalogs/dust/trunk/maps/SFD_dust_4096_sgp.fits"
            downloader(_SFD_URL_SGP, sfd_sgp_path, "SFD (SGP)", test=test)
        return None


import pathlib
import numpy
from numpy.random import default_rng
rng= default_rng()


def test_distance_dependent_forloop():
    # Test that extinction from the Green19 map monotonicaly increases 
    # with distance for a given line of sight.
    from mwdust import Green19
    green19= Green19()
    glons= rng.uniform(50.,220.,size=20)
    glats= rng.uniform(-90.,90.,size=20)
    dists= rng.uniform(0.01,1.,size=5)
    ebvs= numpy.array([[green19(glons[ii],glats[ii],dists[jj]) \
                            for ii in range(len(glons))]
                            for jj in range(len(dists))]).T     
    assert numpy.all(numpy.fabs(ebvs-ebvs[0])>=0), \
        'For-loop Green19 extinction does not monotonically increase with distance for at lease one given line of sight'
    return None


def test_distance_dependent_vectorized():
    # Test that extinction from the Green19 map monotonicaly increases 
    # with distance for a given line of sight.
    from mwdust import Green19
    green19= Green19()
    glons= rng.uniform(50.,220.,size=20)
    glats= rng.uniform(-90.,90.,size=20)
    dists= rng.uniform(0.01,1.,size=5)
    ebvs= numpy.array([[green19(glons[ii],glats[ii],dists) \
                            for ii in range(len(glons))]]).T 
    assert numpy.all(numpy.fabs(ebvs-ebvs[0])>=0), \
        'Vectorized Green19 extinction does not monotonically increase with distance for at lease one given line of sight'
    return None


def test_against_known_values_forloop():
    # Test that the Green19 extinction agrees with a table of known values
    # These were computed using mwdust on Linux using mwdust 1.2
    # before custom healpix functions were implemented (thus computed using healpy)
    from mwdust import Green19
    green19= Green19()
    known= numpy.loadtxt(pathlib.Path(__file__).parent / 'green19_benchmark.dat',delimiter=',')
    glons= known.T[0]
    glats= known.T[1]
    dists= known.T[2]
    ebvs= known.T[3]
    assert numpy.all(ebvs-[green19(glon,glat,dist)[0]
                            for glon,glat,dist in zip(glons,glats,dists)] < 10.**-8.), \
        'For-loop Green19 extinction does not agree with known values'
    return None


def test_against_known_values_vectorized():
    # Test that the Green19 extinction agrees with a table of known values
    # These were computed using mwdust on Linux using mwdust 1.2
    # before custom healpix functions were implemented (thus computed using healpy)
    # and before the vectorized version of the Green19 map was implemented
    from mwdust import Green19
    green19= Green19()
    known= numpy.loadtxt(pathlib.Path(__file__).parent / 'green19_benchmark.dat',delimiter=',')
    glons= known.T[0]
    glats= known.T[1]
    dists= known.T[2]
    ebvs= known.T[3]
    assert numpy.all(ebvs-green19(glons,glats,dists) < 10.**-8.), \
        'Vectorized Green19 extinction does not agree with known values'
    return None



# Test that asking for E(B-V) returns the correct value for all dust maps
import numpy
import mwdust

def test_SFD():
    sfd_ebv= mwdust.SFD(filter='E(B-V)')
    sfd_b= mwdust.SFD(filter='Landolt B')
    sfd_v= mwdust.SFD(filter='Landolt V')
    l,b,d= 10., 1., 2.
    assert numpy.fabs(sfd_ebv(l,b,d)-(sfd_b(l,b,d)-sfd_v(l,b,d))) \
         < 10.**-12., \
        'SFD E(B-V) does not agree with A(B)-A(V)'
    return None

def test_Drimmel03():
    drim_ebv= mwdust.Drimmel03(filter='E(B-V)')
    drim_b= mwdust.Drimmel03(filter='Landolt B')
    drim_v= mwdust.Drimmel03(filter='Landolt V')
    l,b,d= 10., 1., 2.
    assert numpy.fabs(drim_ebv(l,b,d)-(drim_b(l,b,d)-drim_v(l,b,d))) \
         < 10.**-12., \
        'Drimmel03 E(B-V) does not agree with A(B)-A(V)'
    return None

def test_Marshall06():
    mar_ebv= mwdust.Marshall06(filter='E(B-V)')
    mar_b= mwdust.Marshall06(filter='Landolt B')
    mar_v= mwdust.Marshall06(filter='Landolt V')
    l,b,d= 10., 1., 2.
    assert numpy.fabs(mar_ebv(l,b,d)-(mar_b(l,b,d)-mar_v(l,b,d))) \
         < 10.**-12., \
        'Marshall06 E(B-V) does not agree with A(B)-A(V)'
    return None

def test_Green19():
    # Need to preserve memory
    l,b,d= 10., 1., 2.
    green_ebv= mwdust.Green19(filter='E(B-V)')
    ebv= green_ebv(l,b,d)
    del green_ebv
    green_b= mwdust.Green19(filter='Landolt B')
    ab= green_b(l,b,d)
    del green_b
    green_v= mwdust.Green19(filter='Landolt V')
    av= green_v(l,b,d)
    assert numpy.fabs(ebv-(ab-av)) \
         < 10.**-12., \
        'Green19 E(B-V) does not agree with A(B)-A(V)'
    return None

def test_Combined15():
    # Need to preserve memory
    l,b,d= 10., 1., 2.
    combined_ebv= mwdust.Combined15(filter='E(B-V)')
    ebv= combined_ebv(l,b,d)
    del combined_ebv
    combined_b= mwdust.Combined15(filter='Landolt B')
    ab= combined_b(l,b,d)
    del combined_b
    combined_v= mwdust.Combined15(filter='Landolt V')
    av= combined_v(l,b,d)
    assert numpy.fabs(ebv-(ab-av)) \
         < 10.**-12., \
        'Combined15 E(B-V) does not agree with A(B)-A(V)'
    return None



import sys
import distutils.sysconfig as sysconfig
import ctypes
import ctypes.util
from numpy.ctypeslib import ndpointer
import os, os.path
from pathlib import Path
import numpy
import platform
import tqdm
from mwdust.util.download import dust_dir

WIN32= platform.system() == 'Windows'
#Find and load the library
_lib = None
_libname = ctypes.util.find_library('sfd_c')
PY3= sys.version > '3'
if PY3: #pragma: no cover
    _ext_suffix= sysconfig.get_config_var('EXT_SUFFIX')
else:
    _ext_suffix= sysconfig.get_config_var('SO')
if _libname:
    _lib = ctypes.CDLL(_libname)
if _lib is None:
    # Add top-level mwdust repository directory for pip install (-e) .,
    # just becomes site-packages for regular install
    paths = sys.path
    paths.append(str(Path(__file__).parent.parent.parent.absolute()))
    for path in [Path(p) for p in paths]:
        if not path.is_dir():
            continue
        try:
            _lib = ctypes.CDLL(str(path / f"sfd_c{_ext_suffix}"))
        except OSError:
            _lib = None
        else:
            break
if _lib is None:
    raise IOError('SFD/C module not found')

#MAP path names
ebvFileN= os.path.join(dust_dir,'maps','SFD_dust_4096_ngp.fits')
ebvFileS= os.path.join(dust_dir,'maps','SFD_dust_4096_sgp.fits')

def read_SFD_EBV(glon,glat,interp=True,noloop=False,verbose=False,pbar=True):
    """
    NAME:
       read_SFD_EBV
    PURPOSE:
       read an E(B-V) value from the Schlegel, Finkbeiner, & Davis (1998) maps
    INPUT:
       glon - Galactic longitude (deg), can be an array
       glat - Galactic latitude (deg), can be an array
       interp= (True) if True, interpolate using the nearest pixels
       noloop= (False) if True, don't loop through the glons
       verbose= (False) if True, be verbose
       pbar= (True) if True, show progress bar
    OUTPUT:
       array of E(B-V) from Schlegel, Finkbeiner, & Davis (1998)
    HISTORY:
       2013-11-23 - Written - Bovy (IAS)
    """
    #Parse input
    if isinstance(glon,(int,float,numpy.float32,numpy.float64)):
        glon= numpy.array([glon])
    if isinstance(glat,(int,float,numpy.float32,numpy.float64)):
        glat= numpy.array([glat])

    nstar = len(glon)
    if nstar > 1 and pbar:
        pbar= tqdm.tqdm(total=nstar, leave=False)
        pbar_func_ctype= ctypes.CFUNCTYPE(None)
        pbar_c= pbar_func_ctype(pbar.update)
    else: # pragma: no cover
        pbar_c= None

    #Set up the C code
    ndarrayFlags= ('C_CONTIGUOUS','WRITEABLE')
    evalFunc= _lib.lambert_getval
    evalFunc.argtypes= [ctypes.c_char_p,
                        ctypes.c_char_p,
                        ctypes.c_long,
                        ndpointer(dtype=numpy.float32,flags=ndarrayFlags),
                        ndpointer(dtype=numpy.float32,flags=ndarrayFlags),
                        ctypes.c_int,
                        ctypes.c_int,
                        ctypes.c_int, 
                        ndpointer(dtype=numpy.int32,flags=ndarrayFlags),
                        ctypes.c_void_p]
    evalFunc.restype= ctypes.POINTER(ctypes.c_float)

    #Array requirements, first store old order
    f_cont= [glon.flags['F_CONTIGUOUS'],
             glat.flags['F_CONTIGUOUS']]
    glon= numpy.require(glon,dtype=numpy.float64,requirements=['C','W'])
    glat= numpy.require(glat,dtype=numpy.float64,requirements=['C','W'])
    err= numpy.require(0,dtype=numpy.int32,requirements=['C','W'])

    # Check that the filename isn't too long for the SFD code
    if len(ebvFileN.encode('ascii')) >= 120 \
            or len(ebvFileS.encode('ascii')) >= 120:
        raise RuntimeError(f"The path of the file that contains the SFD dust maps is too long ({len(ebvFileN.encode('ascii'))}); please shorten the path of DUST_DIR")

    res= evalFunc(ctypes.c_char_p(ebvFileN.encode('ascii')),
                     ctypes.c_char_p(ebvFileS.encode('ascii')),
                     ctypes.c_long(nstar),
                     glon.astype(numpy.float32,order='C',copy=False),
                     glat.astype(numpy.float32,order='C',copy=False),
                     ctypes.c_int(interp),
                     ctypes.c_int(noloop),
                     ctypes.c_int(verbose), 
                     err,
                     pbar_c
                     )
    if numpy.any(err == -10):
        raise KeyboardInterrupt("Interrupted by CTRL-C (SIGINT)")
    result= numpy.fromiter(res,dtype=float,count=nstar)

    #Reset input arrays
    if f_cont[0]: glon= numpy.asfortranarray(glon)
    if f_cont[1]: glat= numpy.asfortranarray(glat)

    return result



import os
import shutil
from setuptools import setup
from distutils.core import Extension
import sys
import subprocess
import glob
import platform
WIN32= platform.system() == 'Windows'

long_description= ''
previous_line= ''
with open('README.rst') as dfile:
    for line in dfile:
        if not 'image' in line and not 'target' in line \
                and not 'DETAILED' in line and not '**master**' in line \
                and not '**development' in line \
                and not 'DETAILED' in  previous_line:
            long_description+= line
        previous_line= line

#SFD  extension
sfd_c_src= glob.glob('mwdust/util/SFD_CodeC/*.c')

if not WIN32:
    sfd_libraries=['m']
else:
    sfd_libraries=[]
sfd_c= Extension('sfd_c',
                 sources=sfd_c_src,
                 libraries=sfd_libraries,
                 extra_compile_args=['-DLITTLE_ENDIAN'],
                 include_dirs=['mwdust/util/SFD_CodeC'])


#healpix  extension
healpix_c_src= glob.glob('mwdust/util/healpix_CodeC/*.c')

healpix_c= Extension('healpix_c',
                 sources=healpix_c_src,
                 libraries=sfd_libraries,
                 extra_compile_args=['-DLITTLE_ENDIAN'],
                 include_dirs=['mwdust/util/healpix_CodeC'])

ext_modules=[sfd_c, healpix_c]

install_requires= ['numpy','scipy','matplotlib','astropy','h5py','tqdm', 'requests']
if not WIN32:
    install_requires.append('healpy')
setup(name='mwdust',
      version='1.5.post0',
      description='Dust in the Milky Way',
      author='Jo Bovy',
      author_email='bovy@astro.utoronto.ca',
      license='New BSD',
      long_description=long_description,
      url='https://github.com/jobovy/mwdust',
      package_dir = {'mwdust/': ''},
      packages=['mwdust',
                'mwdust/util'],
      package_data={'mwdust/util':['extCurves/extinction.tbl',
                                   'extCurves/apj398709t6_ascii.txt']},
      install_requires=install_requires,
      ext_modules=ext_modules,
      classifiers=[
        "Development Status :: 6 - Mature",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: C",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Astronomy",
        "Topic :: Scientific/Engineering :: Physics"]
      )



###############################################################################
#
#   Drimmel03: extinction model from Drimmel et al. 2003 2003A&A...409..205D
#
###############################################################################
import os
import copy
import numpy
import tarfile
import inspect
from scipy.ndimage import map_coordinates
from scipy import optimize
try:
    import healpy
except ImportError: pass
from mwdust.util.extCurves import aebv
from mwdust.util import read_Drimmel
from mwdust.util.tools import cos_sphere_dist
from mwdust.util.download import  dust_dir, downloader
from mwdust.DustMap3D import DustMap3D

_DEGTORAD= numpy.pi/180.
class Drimmel03(DustMap3D):
    """extinction model from Drimmel et al. 2003 2003A&A...409..205D"""
    def __init__(self,filter=None,sf10=True):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the Drimmel03 dust map
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
        OUTPUT:
           object
        HISTORY:
           2013-12-10 - Started - Bovy (IAS)
        """
        DustMap3D.__init__(self,filter=filter)
        self._sf10= sf10
        #Read the maps
        drimmelMaps= read_Drimmel.readDrimmelAll()
        self._drimmelMaps= drimmelMaps
        #Sines and cosines of sky positions of COBE pixels
        self._rf_sintheta= numpy.sin(numpy.pi/2.-self._drimmelMaps['rf_glat']*_DEGTORAD)
        self._rf_costheta= numpy.cos(numpy.pi/2.-self._drimmelMaps['rf_glat']*_DEGTORAD)
        self._rf_sinphi= numpy.sin(self._drimmelMaps['rf_glon']*_DEGTORAD)
        self._rf_cosphi= numpy.cos(self._drimmelMaps['rf_glon']*_DEGTORAD)
        #Various setups
        self._xsun= -8.
        self._zsun= 0.015
        #Global grids
        self._nx_disk, self._ny_disk, self._nz_disk= 151, 151, 51
        self._dx_disk, self._dy_disk, self._dz_disk= 0.2, 0.2, 0.02
        self._nx_ori, self._ny_ori, self._nz_ori= 76, 151, 51
        self._dx_ori, self._dy_ori, self._dz_ori= 0.05, 0.05, 0.02
        #Local grids
        self._nx_diskloc, self._ny_diskloc, self._nz_diskloc= 31, 31, 51
        self._dx_diskloc, self._dy_diskloc, self._dz_diskloc= 0.05, 0.05, 0.02
        self._nx_ori2, self._ny_ori2, self._nz_ori2= 101, 201, 51
        self._dx_ori2, self._dy_ori2, self._dz_ori2= 0.02, 0.02, 0.02
        return None

    def _evaluate(self,l,b,d,norescale=False,
                  _fd=1.,_fs=1.,_fo=1.):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the dust-map
        INPUT:
           l- Galactic longitude (deg)
           b- Galactic latitude (deg)
           d- distance (kpc) can be array
           norescale= (False) if True, don't apply re-scalings
           _fd, _fs, _fo= (1.) amplitudes of the different components
        OUTPUT:
           extinction
        HISTORY:
           2013-12-10 - Started - Bovy (IAS)
        """
        if isinstance(l,numpy.ndarray) or isinstance(b,numpy.ndarray):
            raise NotImplementedError("array input for l and b for Drimmel dust map not implemented")

        cl= numpy.cos(l*_DEGTORAD)
        sl= numpy.sin(l*_DEGTORAD)
        cb= numpy.cos(b*_DEGTORAD)
        sb= numpy.sin(b*_DEGTORAD)

        #Setup arrays
        avori= numpy.zeros_like(d)
        avspir= numpy.zeros_like(d)
        avdisk= numpy.zeros_like(d)

        #Find nearest pixel in COBE map for the re-scaling
        rfIndx= numpy.argmax(cos_sphere_dist(self._rf_sintheta,
                                             self._rf_costheta,
                                             self._rf_sinphi,
                                             self._rf_cosphi,
                                             numpy.sin(numpy.pi/2.-b*_DEGTORAD),
                                             numpy.cos(numpy.pi/2.-b*_DEGTORAD),
                                             sl,cl))

        rfdisk, rfspir, rfori= 1., 1., 1,
        if self._drimmelMaps['rf_comp'][rfIndx] == 1 and not norescale:
            rfdisk= self._drimmelMaps['rf'][rfIndx]
        elif self._drimmelMaps['rf_comp'][rfIndx] == 2 and not norescale:
            rfspir= self._drimmelMaps['rf'][rfIndx]
        elif self._drimmelMaps['rf_comp'][rfIndx] == 3 and not norescale:
            rfori= self._drimmelMaps['rf'][rfIndx]

        #Find maximum distance
        dmax= 100.
        if b != 0.: dmax= .49999/numpy.fabs(sb) - self._zsun/sb
        if cl != 0.:
            tdmax= (14.9999/numpy.fabs(cl)-self._xsun/cl)
            if tdmax < dmax: dmax= tdmax
        if sl != 0.:
            tdmax = 14.9999/numpy.fabs(sl)
            if tdmax < dmax: dmax= tdmax
        d= copy.copy(d)
        d[d > dmax]= dmax
        
        #Rectangular coordinates
        X= d*cb*cl
        Y= d*cb*sl
        Z= d*sb+self._zsun

        #Local grid
        #Orion
        locIndx= (numpy.fabs(X) < 1.)*(numpy.fabs(Y) < 2.)
        if numpy.sum(locIndx) > 0:
            xi = X[locIndx]/self._dx_ori2+float(self._nx_ori2-1)/2.
            yj = Y[locIndx]/self._dy_ori2+float(self._ny_ori2-1)/2.
            zk = Z[locIndx]/self._dz_ori2+float(self._nz_ori2-1)/2.
            avori[locIndx]= map_coordinates(self._drimmelMaps['avori2'],
                                            [xi,yj,zk],
                                            mode='constant',cval=0.)
        #local disk
        locIndx= (numpy.fabs(X) < 0.75)*(numpy.fabs(Y) < 0.75)
        if numpy.sum(locIndx) > 0:
            xi = X[locIndx]/self._dx_diskloc+float(self._nx_diskloc-1)/2.
            yj = Y[locIndx]/self._dy_diskloc+float(self._ny_diskloc-1)/2.
            zk = Z[locIndx]/self._dz_diskloc+float(self._nz_diskloc-1)/2.
            avdisk[locIndx]= map_coordinates(self._drimmelMaps['avdloc'],
                                             [xi,yj,zk],
                                             mode='constant',cval=0.)
        
        #Go to Galactocentric coordinates
        X= X+self._xsun

        #Stars beyond the local grid
        #Orion
        globIndx= True^(numpy.fabs(X-self._xsun) < 1.)*(numpy.fabs(Y) < 2.)
        if numpy.sum(globIndx) > 0:
            #Orion grid is different from other global grids, so has its own dmax
            dmax= 100.
            if b != 0.: dmax= .49999/numpy.fabs(sb) - self._zsun/sb
            if cl > 0.:
                tdmax = (2.374999/numpy.fabs(cl))
                if tdmax < dmax: dmax= tdmax
            if cl < 0.:
                tdmax = (1.374999/numpy.fabs(cl))
                if tdmax < dmax: dmax= tdmax
            if sl != 0.:
                tdmax = (3.749999/numpy.fabs(sl))
                if tdmax < dmax: dmax= tdmax
            dori= copy.copy(d)
            dori[dori > dmax]= dmax
            Xori= dori*cb*cl+self._xsun
            Yori= dori*cb*sl
            Zori= dori*sb+self._zsun

            xi = Xori[globIndx]/self._dx_ori + 2.5*float(self._nx_ori-1)
            yj = Yori[globIndx]/self._dy_ori + float(self._ny_ori-1)/2.
            zk = Zori[globIndx]/self._dz_ori + float(self._nz_ori-1)/2.

            avori[globIndx]= map_coordinates(self._drimmelMaps['avori'],
                                             [xi,yj,zk],
                                             mode='constant',cval=0.)
        #disk & spir
        xi = X/self._dx_disk+float(self._nx_disk-1)/2.
        yj = Y/self._dy_disk+float(self._ny_disk-1)/2.
        zk = Z/self._dz_disk+float(self._nz_disk-1)/2.
        avspir= map_coordinates(self._drimmelMaps['avspir'],
                                [xi,yj,zk],
                                mode='constant',cval=0.)
        globIndx= True^(numpy.fabs(X-self._xsun) < 0.75)*(numpy.fabs(Y) < 0.75)
        if numpy.sum(globIndx) > 0:
            avdisk[globIndx]= map_coordinates(self._drimmelMaps['avdisk'],
                                              [xi,yj,zk],
                                              mode='constant',
                                              cval=0.)[globIndx]
        
        #Return
        out=_fd*rfdisk*avdisk+_fs*rfspir*avspir+_fo*rfori*avori
        if self._filter is None: # From Rieke & Lebovksy (1985); if sf10, first put ebv on SFD scale
            return out/3.09/((1-self._sf10)+self._sf10*0.86)
        else: 
            return out/3.09/((1-self._sf10)+self._sf10*0.86)\
                *aebv(self._filter,sf10=self._sf10)

    def dust_vals_disk(self,lcen,bcen,dist,radius):
        """
        NAME:
           dust_vals_disk
        PURPOSE:
           return the distribution of extinction within a small disk as samples
        INPUT:
           lcen, bcen - Galactic longitude and latitude of the center of the disk (deg)
           dist - distance in kpc
           radius - radius of the disk (deg)
        OUTPUT:
           (pixarea,extinction) - arrays of pixel-area in sq rad and extinction value
        HISTORY:
           2015-03-07 - Written - Bovy (IAS)
        """
        # Convert the disk center to a HEALPIX vector
        vec= healpy.pixelfunc.ang2vec((90.-bcen)*_DEGTORAD,lcen*_DEGTORAD)
        # We pixelize the map with a HEALPIX grid with nside=256, to somewhat
        # oversample the Drimmel resolution
        nside= 256
        # Find the pixels at this resolution that fall within the disk
        ipixs= healpy.query_disc(nside,vec,radius*_DEGTORAD,
                                 inclusive=False,nest=False)
        # Query the HEALPIX map for pixels that lie within the disk
        pixarea= healpy.pixelfunc.nside2pixarea(nside)+numpy.zeros(len(ipixs))
        extinction= []
        for ii, ipix in enumerate(ipixs):
            # Get glon and glat
            b9, l= healpy.pixelfunc.pix2ang(nside,ipix,nest=False)
            b= 90.-b9/_DEGTORAD
            l/= _DEGTORAD
            # Now evaluate
            extinction.append(self._evaluate(l,b,dist))
        extinction= numpy.array(extinction)
        return (pixarea,extinction)

    def fit(self,l,b,dist,ext,e_ext):
        """
        NAME:
           fit
        PURPOSE:
           fit the amplitudes of the disk, spiral, and Orion parts of the
           Drimmel map to other data
        INPUT:
           l,b- Galactic longitude and latitude in degree
           dist - distance in kpc
           ext - extinction at dist
           e_ext - error in extinction
        OUTPUT:
           (fd,fs,fo,dist_stretch) amplitudes of disk, spiral, and Orion parts
           and a 'distance stretch' applied to the model
           (applied as self(l,b,dist*dist_stretch))
        HISTORY:
           2013-12-16 - Written - Bovy (IAS)
        """
        #Fit consists of 
        #a) overall amplitude A
        #b) relative amplitude fd/A, fs/A
        #c) distance stretch
        pars= numpy.array([0.,numpy.log(1./3.),numpy.log(1./3.),0.])
        pars=\
            optimize.fmin_powell(_fitFunc,pars,args=(self,l,b,dist,ext,e_ext))
        amp= numpy.exp(pars[0])
        fd= amp*numpy.exp(pars[1])
        fs= amp*numpy.exp(pars[2])
        fo= amp*(1.-fd-fs)
        return (fd,fs,fo,numpy.exp(pars[3]))

    @classmethod
    def download(cls, test=False):
        drimmel_folder_path = os.path.abspath(os.path.join(inspect.getfile(cls), "..", "util", "drimmeldata"))
        drimmel_path = os.path.join(drimmel_folder_path, "data-for.tar.gz")
        if not os.path.exists(drimmel_path):
            if not os.path.exists(drimmel_folder_path):
                os.mkdir(drimmel_folder_path)
            _DRIMMEL_URL= "https://zenodo.org/record/7340108/files/data-for.tar.gz"
            downloader(_DRIMMEL_URL, drimmel_path, cls.__name__, test=test)
            if not test:
                drim_file = tarfile.open(drimmel_path)
                drim_file.extractall(drimmel_folder_path)
                drim_file.close()
        return None

def _fitFunc(pars,drim,l,b,dist,ext,e_ext):
    amp= numpy.exp(pars[0])
    fd= amp*numpy.exp(pars[1])
    fs= amp*numpy.exp(pars[2])
    fo= amp*(1.-fd-fs)
    dist_stretch= numpy.exp(pars[3])
    model_ext= drim(l,b,dist*dist_stretch,_fd=fd,_fs=fs,_fo=fo)
    return 0.5*numpy.sum((model_ext-ext)**2./e_ext**2.)



###############################################################################
#
#    mwutils.util.tools: some general tools
#
###############################################################################


def cos_sphere_dist(sintheta,costheta,
                    sinphi,cosphi,
                    sintheta_o,costheta_o,
                    sinphi_o,cosphi_o):
    """
    NAME:
       cos_sphere_dist
    PURPOSE:
       computes the cosine of the spherical distance between two
       points on the sphere
    INPUT:
       theta  - polar angle [0,pi]
       phi    - azimuth [0,2pi]
       theta_o  - polar angle of center of the disk
       phi_o  - azimuth of the center of the disk
    OUTPUT:
       spherical distance
    HISTORY:
       2010-04-29 -Written - Bovy (NYU)
    """
    return (sintheta*sintheta_o
            *(cosphi_o*cosphi+
              sinphi_o*sinphi)+
            costheta_o*costheta)



###############################################################################
#
#   Combined15: extinction model obtained from a combination of Marshall et al.
#               (2006), Green et al. (2015), and Drimmel et al. (2003)
#
###############################################################################
import os, os.path
import numpy
import h5py
from mwdust.util.download import downloader, dust_dir
from mwdust.HierarchicalHealpixMap import HierarchicalHealpixMap
_DEGTORAD= numpy.pi/180.
_combineddir= os.path.join(dust_dir, 'combined15')
class Combined15(HierarchicalHealpixMap):
    """extinction model obtained from a combination of Marshall et al.
    (2006), Green et al. (2015), and Drimmel et al. (2003)"""
    def __init__(self,filter=None,sf10=True,load_samples=False,
                 interpk=1):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the combined dust map
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
           interpk= (1) interpolation order
        OUTPUT:
           object
        HISTORY:
           2015-07-28 - Started - Bovy (UofT)
        """
        HierarchicalHealpixMap.__init__(self,filter=filter,sf10=sf10)
        #Read the map
        with h5py.File(os.path.join(_combineddir,'dust-map-3d.h5'),'r') \
                as combineddata:
            self._pix_info= combineddata['/pixel_info'][:]
            self._best_fit= combineddata['/best_fit'][:]
        # Utilities
        self._distmods= numpy.linspace(4.,19.,31)
        self._minnside= numpy.amin(self._pix_info['nside'])
        self._maxnside= numpy.amax(self._pix_info['nside'])
        nlevels= int(numpy.log2(self._maxnside//self._minnside))+1
        self._nsides= [self._maxnside//2**ii for ii in range(nlevels)]
        self._indexArray= numpy.arange(len(self._pix_info['healpix_index']))
        # For the interpolation
        self._intps= numpy.zeros(len(self._pix_info['healpix_index']),
                                 dtype='object') #array to cache interpolated extinctions
        self._interpk= interpk
        return None
    
    @classmethod
    def download(cls, test=False):
        # Download the combined map of Bovy et al. (2015): Marshall+Green+Drimmel for full sky coverage
        combined15_path = os.path.join(dust_dir, "combined15", "dust-map-3d.h5")
        if not os.path.exists(combined15_path):
                if not os.path.exists(os.path.join(dust_dir, "combined15")):
                    os.mkdir(os.path.join(dust_dir, "combined15"))
                _COMBINED15_URL = "https://zenodo.org/record/31262/files/dust-map-3d.h5"
                downloader(_COMBINED15_URL, combined15_path, cls.__name__, test=test)
        return None


###############################################################################
#
#   HierarchicalHealpixMap: General class for extinction maps given as
#                           a hierarchical HEALPix pixelation (e.g., Green
#                           et al. 2015)
#
###############################################################################
import numpy
from scipy import interpolate
from mwdust.util.healpix import ang2pix
from mwdust.util.extCurves import aebv
from mwdust.DustMap3D import DustMap3D
_DEGTORAD= numpy.pi/180.
class HierarchicalHealpixMap(DustMap3D):
    """General class for extinction maps given as a hierarchical HEALPix 
    pixelation (e.g., Green et al. 2015) """
    def __init__(self,filter=None,sf10=True):
        """
        NAME:
           __init__
        PURPOSE:
           Initialize the combined dust map
        INPUT:
           filter= filter to return the extinction in
           sf10= (True) if True, use the Schlafly & Finkbeiner calibrations
        OUTPUT:
           object
        HISTORY:
           2015-07-28 - Started - Bovy (UofT)
        """
        DustMap3D.__init__(self,filter=filter)
        self._sf10= sf10
        return None


    def _evaluate(self, ls, bs, ds):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the dust-map for array input
        INPUT:
           l- Galactic longitude (deg) can be array
           b- Galactic latitude (deg) can be array
           d- distance (kpc) can be array
        OUTPUT:
           extinction E(B-V)
        HISTORY:
           2015-03-02 - Started - Bovy (IAS)
           2023-07-05 - Vectorized - Henry Leung (UofT)
        """
        ls = numpy.atleast_1d(ls)
        bs = numpy.atleast_1d(bs)
        ds = numpy.atleast_1d(ds)

        distmod= 5.*numpy.log10(ds)+10.
        lbIndx= self._lbIndx(ls, bs)
        if len(ls) == 1 and len(ds) > 1:
            lbIndx = numpy.tile(lbIndx, len(ds))

        result = numpy.zeros_like(ds)
        for counter, i, d in zip(numpy.arange(len(result)), lbIndx, distmod):
            if self._intps[i] != 0:
                out= self._intps[i](d)
            else:
                interpData=\
                    interpolate.InterpolatedUnivariateSpline(self._distmods,
                                                            self._best_fit[i],
                                                            k=self._interpk)
                out= interpData(d)
                self._intps[i]= interpData
            result[counter] = out
        if self._filter is not None:
            result =  result * aebv(self._filter,sf10=self._sf10)
        # set nan for invalid indices
        result[lbIndx==-1] = numpy.nan
        return result


    def dust_vals_disk(self,lcen,bcen,dist,radius):
        """
        NAME:
           dust_vals_disk
        PURPOSE:
           return the distribution of extinction within a small disk as samples
        INPUT:
           lcen, bcen - Galactic longitude and latitude of the center of the disk (deg)
           dist - distance in kpc
           radius - radius of the disk (deg)
        OUTPUT:
           (pixarea,extinction) - arrays of pixel-area in sq rad and extinction value
        HISTORY:
           2015-03-06 - Written - Bovy (IAS)
        """
        try:
            import healpy
        except ImportError:
            raise ModuleNotFoundError("This function requires healpy to be installed")
        # Convert the disk center to a HEALPIX vector
        vec= healpy.pixelfunc.ang2vec((90.-bcen)*_DEGTORAD,lcen*_DEGTORAD)
        distmod= 5.*numpy.log10(dist)+10.
        # Query the HEALPIX map for pixels that lie within the disk
        pixarea= []
        extinction= []
        for nside in self._nsides:
            # Find the pixels at this resolution that fall within the disk
            ipixs= healpy.query_disc(nside,vec,radius*_DEGTORAD,
                                    inclusive=False,nest=True)
            # Get indices of all pixels within the disk at current nside level
            nsideindx= self._pix_info['nside'] == nside
            potenIndxs= self._indexArray[nsideindx]
            nsidepix= self._pix_info['healpix_index'][nsideindx]
            # Loop through the pixels in the (small) disk
            tout= []
            for ii,ipix in enumerate(ipixs):
                lbIndx= potenIndxs[ipix == nsidepix]
                if numpy.sum(lbIndx) == 0: continue
                if self._intps[lbIndx] != 0:
                    tout.append(self._intps[lbIndx][0](distmod))
                else:
                    interpData=\
                        interpolate.InterpolatedUnivariateSpline(self._distmods,
                                                                 self._best_fit[lbIndx],
                                                                 k=self._interpk)
                    tout.append(interpData(distmod))
                    self._intps[lbIndx]= interpData
            tarea= healpy.pixelfunc.nside2pixarea(nside)
            tarea= [tarea for ii in range(len(tout))]
            pixarea.extend(tarea)
            extinction.extend(tout)
        pixarea= numpy.array(pixarea)
        extinction= numpy.array(extinction)
        if not self._filter is None:
            extinction= extinction*aebv(self._filter,sf10=self._sf10)        
        return (pixarea,extinction)

    def _lbIndx(self, ls, bs):
        """Return the indices in the _combineddata array corresponding to arrays of (l, b)"""
        stop_mask = numpy.zeros(len(ls), dtype=bool)  # mask to be accumulated when looping through nside
        indx_result = numpy.ones(len(ls), dtype=int) * -1  # -1 for bad star, int array has no nan
        for nside in self._nsides:
            tpix = ang2pix(nside, (90.-bs)*_DEGTORAD, ls*_DEGTORAD, nest=True)
            nside_idx = numpy.where(self._pix_info['nside'] == nside)[0]
            healpix_index_nside = self._pix_info['healpix_index'][nside_idx]
            sorted_order = numpy.argsort(healpix_index_nside)
            # use searchsorted to find the index of tpix in healpix_index_nside efficiently
            result = numpy.searchsorted(healpix_index_nside, tpix, sorter=sorted_order)
            # need to deal with indices where they are larger than the largest healpix_index_nside
            known_bad_idx = (result == len(nside_idx))
            result[known_bad_idx] = 0  # wrap around
            result = sorted_order[result]  # reverse the sorting before indexing
            result = nside_idx[result]
            good_result_idx = ((self._pix_info['healpix_index'][result] == tpix) & (self._pix_info['nside'][result] == nside) & (~known_bad_idx))
            indx_result = numpy.where(~stop_mask & good_result_idx, result, indx_result)
            indx_result = numpy.where(known_bad_idx & ~stop_mask, -1, indx_result)  # set bad star to -1
            stop_mask = stop_mask | good_result_idx  # update mask for the next nside
        return indx_result

    def plot_mollweide(self,d,**kwargs):
        """
        NAME:
           plot_mollweide
        PURPOSE:
           plot the extinction across the sky in Galactic coordinates  out to a given distance using a Mollweide projection
        INPUT:
           d - distance in kpc (nearest distance to this in the map is plotted)
           nside_plot= (2048) nside of the plotted map
           healpy.visufunc.mollview kwargs
        OUTPUT:
           plot to output device
        HISTORY:
           2019-12-06 - Written - Bovy (UofT)
        """
        try:
            import healpy
        except ImportError:
            raise ModuleNotFoundError("This function requires healpy to be installed")
        # Distance modulus
        dm= 5.*numpy.log10(d)+10.
        # Get factor to apply to map to obtain extinction in object's filter
        filter_fac= aebv(self._filter,sf10=self._sf10) \
                    if not self._filter is None else 1.
        # Map the dust map to a common nside, first find nearest distance pixel
        tpix= numpy.argmin(numpy.fabs(dm-self._distmods))
        # Construct an empty map at the highest HEALPix resolution present in the map; code snippets adapted from http://argonaut.skymaps.info/usage
        nside_max= numpy.max(self._pix_info['nside'])
        npix= healpy.pixelfunc.nside2npix(nside_max)
        pix_val= numpy.empty(npix,dtype='f8')
        pix_val[:] = healpy.UNSEEN
        # Fill the upsampled map
        for nside in numpy.unique(self._pix_info['nside']):
            # Get indices of all pixels at current nside level
            indx= self._pix_info['nside'] == nside
            # Extract A_X of each selected pixel
            pix_val_n= filter_fac*self._best_fit[indx,tpix]
            # Determine nested index of each selected pixel in upsampled map
            mult_factor = (nside_max//nside)**2
            pix_idx_n = self._pix_info['healpix_index'][indx]*mult_factor
            # Write the selected pixels into the upsampled map
            for offset in range(mult_factor):
                pix_val[pix_idx_n+offset] = pix_val_n[:]
        # If the desired nside is less than the maximum nside in the map, degrade
        nside_plot= kwargs.get('nside_plot',2048)
        if not nside_plot is None and nside_plot < nside_max:
            pix_val= healpy.pixelfunc.ud_grade(pix_val,
                                               nside_plot,pess=False,
                                               order_in='NEST', 
                                               order_out='NEST')
        pix_val[pix_val == healpy.UNSEEN]= -1.

        if not self._filter is None:
            kwargs['unit']= r'$A_{%s}\,(\mathrm{mag})$' % (self._filter.split(' ')[-1])
        else:
            kwargs['unit']= r'$E(B-V)\,(\mathrm{mag})$'
        kwargs['title']= kwargs.get('title',"")
        healpy.visufunc.mollview(pix_val,
                                 nest=True,
                                 xsize=4000,
                                 min=0.,
                                 max=numpy.quantile(pix_val,0.99),
                                 format=r'$%g$',
                                 cmap='gist_yarg',
                                 **kwargs)
        return None